"""Symbolic computation of minimal cover with orthotopes.

References
==========

Olivier Coudert
    "Two-level logic minimization: An overview"
    Integration, the VLSI Journal
    Vol.17, No.2, Oct 1994, pp.97--140
    http://dx.doi.org/10.1016/0167-9260(94)00007-7
"""
# Copyright 2016 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
from __future__ import absolute_import
from __future__ import print_function
from itertools import cycle
import logging
import time

import natsort
from omega.logic import syntax as stx
from omega.symbolic.bdd import support_issubset
from omega.symbolic.bdd import joint_support
# import polytope (inline)


log = logging.getLogger(__name__)


VAR_OWNER = 'other'


def minimize(f, care, fol):
    """Compute minimal DNF of predicate `f` over integers.

    @param f: predicate over integer-valued variables
    @param care: care set as predicate over same variables
    @type f, care: BDD node
    @type fol: `omega.symbolic.fol.Context`

    @return: minimal cover as BDD over parameters
    @rtype: BDD node
    """
    # reasons for permisiveness here:
    #
    # - enable inspecting env violations of assumption
    # - make errors visible
    # - use entire instantiation domain
    # - permit computing DNF for care set using same `fol.vars`
    # - tests
    if not _care_implies_type_hints(f, care, fol):
        log.warning('care set should imply type hints')
    # could let
    #     f &= care
    # but explicit is better.
    # Also, this permits working outside type hints.
    if not _f_implies_care(f, care, fol):
        log.warning('f should imply care set')
    if (f | ~ care) == fol.true:
        log.warning('f covers care set, so trivial cover')
    log.info('---- branching ----')
    path_cost = 0.0
    x_vars, px, qx, p_to_q = _setup_aux_vars(f, care, fol)
    u_leq_p, p_leq_u = _partial_order(px, fol)
    varmap = _parameter_varmap(px, qx)
    p_leq_q = _orthotope_subseteq(varmap, fol)
    p_eq_q = _orthotope_eq(varmap, fol)
    # covering problem
    fcare = f | ~ care
    # the slack is introduced by having more primes
    # (those for `fcare`) to cover the same minterms (`f`)
    x = _embed_as_implicants(f, px, fol)
    y = prime_orthotopes(
        fcare, px, qx,
        p_leq_q, p_eq_q,
        fol, x_vars)
    bab = _BranchAndBound(
        p_leq_q, p_leq_u, u_leq_p,
        p_eq_q, p_to_q, px, qx, fol)
    # initialize upper bound
    bab.upper_bound = _upper_bound(
        x, y, p_leq_q, p_to_q, fol)
    # assert covers(bab.best_cover, f, p_leq_q, p_to_q, px, fol)
    cover, _ = _traverse(x, y, path_cost, bab, fol)
    if cover is None:
        cover = _some_cover(x, y, p_leq_q, p_to_q, fol)
    assert cover is not None
    assert _covers(cover, f, p_leq_q, p_to_q, px, fol)
    low = care & ~ f
    assert _none_covered(cover, low, p_to_q, px, qx, fol)
    log.info('==== branching ==== ')
    return cover


def _traverse(x, y, path_cost, bab, fol):
    """Compute cyclic core and terminate, prune, or recurse."""
    log.info('\n\n---- traverse ----')
    t0 = time.time()
    xcore, ycore, essential = _cyclic_core_fixpoint(
        x, y, bab, fol)
    _print_cyclic_core(
        x, y, xcore, ycore, essential,
        t0, bab.px, fol)
    # C_left.path =
    #     C.path + 1  (* already from `_branch` *)
    # C_left.lower =
    #     + Cardinality(essential_left)  (* `_cost` *)
    #     + LowerBound(core_left)  (* `lb` below *)
    #
    # C_right.path =
    #     C.path  (* already from `_branch` *)
    # C_right.lower =
    #     + Cardinality(essential_right)
    #     + LowerBound(core_right)
    cost_ess = _cost(essential, bab.p_to_q, fol)
    core_lb = _lower_bound(
        xcore, ycore, bab.p_leq_q, bab.p_to_q, fol)
    sub_lb = cost_ess + core_lb
    branch_lb = path_cost + sub_lb
    if xcore == fol.false:
        assert core_lb == 0, core_lb
        bab.upper_bound = branch_lb
        log.info('terminal case (empty cyclic core)\n'
                 '==== traverse ====\n')
        return essential, sub_lb
    # set global lower bound only once at the top
    # because farther below in the search tree the
    # lower bounds are local, not global
    if bab.lower_bound is None:
        log.info(
            'global lower bound: {lb}'.format(lb=branch_lb))
        bab.lower_bound = branch_lb
    # prune
    # C_left.path + C_left.lower >= global_upper_bound ?
    # C_right.path + C_right.lower >= global_upper_bound ?
    if branch_lb >= bab.upper_bound:
        log.info('prune\n==== traverse ====\n')
        return None, sub_lb
    assert xcore != fol.false
    assert ycore != fol.false
    # branch
    longer_path_cost = path_cost + cost_ess
    r = _branch(xcore, ycore, longer_path_cost, bab, fol)
    # both branches pruned ?
    if r is None:
        log.info('both branches pruned\n'
                 '==== traverse ====\n')
        return None, sub_lb
    cover = r | essential
    log.info('==== traverse ====\n')
    return cover, sub_lb


def _branch(x, y, path_cost, bab, fol):
    log.info('\n\n---- branch ----')
    d = fol.pick(y)
    log.info('picked branching y: {d}'.format(d=d))
    y_branch = fol.assign_from(d)
    ynew = y & ~ y_branch
    assert ynew != y
    # r(p) == p <= y_branch
    dq = {bab.p_to_q[k]: v for k, v in d.items()}
    r = fol.let(dq, bab.p_leq_q)  # those x under y_branch
    x_minus_y = x & ~ r
    assert x_minus_y != x  # must prove always the case
    log.info('left branch')
    e0, left_lb = _traverse(
        x_minus_y, ynew, path_cost + 1, bab, fol)
    # pruning with left lower bound (Thm.7 [Coudert 1994])
    if path_cost + left_lb >= bab.upper_bound:
        log.info(
            'prune both left and right branches\n'
            '==== branch ====\n')
        assert e0 is None, e0
        return None
    log.info('right branch')
    e1, _ = _traverse(x, ynew, path_cost, bab, fol)
    # pick cheaper
    cost_0 = _cost(e0, bab.p_vars, fol)
    cost_1 = _cost(e1, bab.p_vars, fol)
    if cost_0 < cost_1:
        # can be reached only if `e0 != None`
        assert e0 is not None, e0
        e = e0 | y_branch
    else:
        e = e1
    log.info('==== branch ====\n')
    return e


def _cost(u, p_vars, fol):
    """Return numerical cost of cover `u`."""
    if u is None:
        return float('inf')
    # cost of each implicant = 1
    # cost of a cover = number of implicants it contains
    assert support_issubset(u, p_vars, fol)
    n = fol.count(u)
    return n


def cyclic_core(f, care, fol):
    """Shallow minimal cover, only up to cyclic core."""
    log.info('cyclic core computation')
    t0 = time.time()
    # assert
    assert f in fol.bdd, f
    assert care in fol.bdd, care
    assert care != fol.false, 'empty care set'
    assert f != fol.false, 'nothing to cover'
    assert f != fol.true or care != fol.true, (
        'no variables involved in problem')
    x_vars, px, qx, p_to_q = _setup_aux_vars(f, care, fol)
    fcare = ~ care | f
    u_leq_p, p_leq_u = _partial_order(px, fol)
    varmap = _parameter_varmap(px, qx)
    p_leq_q = _orthotope_subseteq(varmap, fol)
    p_eq_q = _orthotope_eq(varmap, fol)
    bab = _BranchAndBound(
        p_leq_q, p_leq_u, u_leq_p, p_eq_q,
        p_to_q, px, qx, fol)
    # covering problem
    x = _embed_as_implicants(f, px, fol)
    y = prime_orthotopes(
        fcare, px, qx,
        p_leq_q, p_eq_q,
        fol, x_vars)
    # assert problem is feasible
    assert x != fol.false
    assert y != fol.false
    assert _covers(y, f, p_leq_q, p_to_q, px, fol)
    xcore, ycore, essential = _cyclic_core_fixpoint(
        x, y, bab, fol)
    if xcore == fol.false:
        assert _covers(essential, f, p_leq_q, p_to_q, px, fol)
    _print_cyclic_core(
        x, y, xcore, ycore, essential,
        t0, px, fol)
    s = dumps_cover(essential, f, care, fol)
    log.info(s)
    return xcore, ycore, essential


def _print_cyclic_core(
        x, y, xcore, ycore, essential,
        t0, px, fol):
    """Print results of cyclic core computation.

    Assert support and covering properties.
    """
    if log.getEffectiveLevel() > logging.INFO:
        return
    # assert
    params = _collect_parameters(px)
    if essential != fol.false:
        assert support_issubset(essential, params, fol)
    if xcore != fol.false:
        assert support_issubset(xcore, params, fol)
    if ycore != fol.false:
        assert support_issubset(ycore, params, fol)
    # print
    m = fol.count(x)
    n = fol.count(y)
    log.info((
        '(x={m}, y={n}) implicants of '
        'covering problem').format(
            m=m, n=n))
    m = fol.count(xcore)
    n = fol.count(ycore)
    log.info((
        '(x={m}, y={n}) implicants after '
        'removing essential elements').format(
            m=m, n=n))
    n = fol.count(essential)
    log.info('{n} primes are essential'.format(n=n))
    t1 = time.time()
    dt = t1 - t0
    log.info('cyclic core took {dt:1.2f} sec'.format(dt=dt))


def _cyclic_core_fixpoint(x, y, bab, fol):
    """Return cyclic core and essential elements."""
    log.debug('\n\n---- cyclic core fixpoint ----')
    assert x in fol.bdd, x
    assert y in fol.bdd, y
    assert support_issubset(x, bab.p_vars, fol)
    assert support_issubset(y, bab.p_vars, fol)
    # compute
    essential = fol.false
    xold, yold = None, None
    i = 0
    while x != xold or y != yold:
        log.debug('starting iteration {i}'.format(i=i))
        t0 = time.time()
        xold, yold = x, y
        x = _max_transpose(x, y, bab, fol, signatures=True)
        e = x & y
        x = x & ~ e
        y = y & ~ e
        essential |= e
        y = _max_transpose(x, y, bab, fol)
        t1 = time.time()
        dt = t1 - t0
        log.debug('iteration {i} took {dt:1.2f} sec'.format(
            i=i, dt=dt))
        i += 1
    _assert_fixpoint(
        x, y, xold, yold, essential, bab.p_vars, fol)
    log.debug('==== cyclic core fixpoint ====\n')
    return x, y, essential


def _assert_fixpoint(x, y, xold, yold, essential, pvars, fol):
    """Assert `x, y = xold, yold` and check supports."""
    assert x == xold
    assert y == yold
    e = x & y
    assert e == fol.false, e
    e = x & essential
    assert e == fol.false, e
    e = y & essential
    assert e == fol.false, e
    assert support_issubset(x, pvars, fol)
    assert support_issubset(y, pvars, fol)
    assert support_issubset(essential, pvars, fol)


def _max_transpose(p_is_signature, p_is_prime,
                   bab, fol, signatures=False):
    """Maximal transposed primes or signatures.

    (max tau_Y(X) or max tau_X(Y))

    @param signatures: if `True`, then transpose signatures,
        otherwise primes.

    Requires that `p, p', q` be in `fol.vars` and
    be refined by the same number of bits each.
    """
    log.info('---- max transpose ----')
    assert support_issubset(p_is_prime, bab.p_vars, fol)
    assert support_issubset(p_is_signature, bab.p_vars, fol)
    # compute
    u = _floor(
        p_is_signature, p_is_prime,
        bab, fol, signatures=signatures)
    r = _maxima(u, bab, fol)
    assert support_issubset(r, bab.p_vars, fol)
    log.info('==== max transpose ====')
    return r


def _floor(p_is_signature, p_is_prime,
           bab, fol, signatures=False):
    """Transpose primes (tau_X(Y)) or signatures (tau_Y(X)).

    @param p_is_prime: some primes, function of `p`
    @param p_is_signature: function of `p`
    @param signatures: if `True`, then transpose signatures,
        otherwise primes.
    """
    log.info('---- tranpose ----')
    p_leq_u = bab.p_leq_u
    u_leq_p = bab.u_leq_p
    if signatures:
        p_is_signature, p_is_prime = p_is_prime, p_is_signature
        u_leq_p, p_leq_u = p_leq_u, u_leq_p
    # assert
    assert support_issubset(p_is_prime, bab.p_vars, fol)
    assert support_issubset(p_is_signature, bab.p_vars, fol)
    # p' used as u
    u_is_signature = fol.let(bab.p_to_u, p_is_signature)
    # same coverage
    p_like_q = _contains_covered(
        u_is_signature, u_leq_p, bab, fol)
    u_like_q = fol.let(bab.p_to_u, p_like_q)
    # u_is_prime = fol.let(p_to_u, p_is_prime)
    q_is_prime = fol.let(bab.p_to_q, p_is_prime)
    #
    r = ~ u_like_q | p_leq_u
    r = fol.forall(bab.u_vars, r)
    r &= p_like_q
    r &= q_is_prime
    r = fol.exist(bab.q_vars, r)
    '''
    q = ', '.join(iter(p_to_q.values()))
    u = ', '.join(iter(p_to_u.values()))
    s = (
        '\E {q}:  {q_is_prime} /\ {p_like_q} /\ '
        '    \A {u}:  {u_like_q} => {p_leq_u}').format(
            q=q,
            u=u,
            q_is_prime=q_is_prime,
            p_like_q=p_like_q,
            u_like_q=u_like_q,
            p_leq_u=p_leq_u)
    r = fol.add_expr(s)
    '''
    assert support_issubset(r, bab.p_vars, fol)
    log.info('==== tranpose ====')
    return r


def _contains_covered(u_is_signature, u_leq_p, bab, fol):
    """Return primes that cover all signatures under prime.

    Require that `p_to_u` be given explicitly,
    to avoid errors if support is empty.

    In the proof, this operator is equivalent to:
        `IsAbove(p, ThoseUnder(u_is_signature, q, Leq))`

    @param signatures: function of `u`
    """
    log.info('---- contains covered ----')
    # assert
    pq_vars = bab.p_vars.union(bab.q_vars)
    pu_vars = bab.p_vars.union(bab.u_vars)
    assert support_issubset(u_is_signature, bab.u_vars, fol)
    assert support_issubset(u_leq_p, pu_vars, fol)
    # compute
    u_leq_q = fol.let(bab.p_to_q, u_leq_p)
    r = u_is_signature & u_leq_q
    r = ~ r | u_leq_p
    r = fol.forall(bab.u_vars, r)
    '''
    uvars = ', '.join(bab.u_vars)
    s = (
        '\A {uvars}:  '
        '    ({sig_u} /\ {u_leq_q}) '
        '        => {u_leq_p}').format(
        uvars=uvars,
        sig_u=u_is_signature,
        u_leq_q=u_leq_q,
        u_leq_p=u_leq_p)
    r = fol.add_expr(s)
    '''
    assert support_issubset(r, pq_vars, fol)
    log.info('==== contains covered ====')
    return r


def _maxima(u, bab, fol):
    """Return maximal elements of `u` wrt `bab.p_leq_q`.

    @param u: function of `bab.p_vars`
    """
    assert support_issubset(u, bab.p_vars, fol)
    v = fol.let(bab.p_to_q, u)
    #
    r = v & bab.p_leq_q
    r = ~ r | bab.p_eq_q
    r = fol.forall(bab.q_vars, r)
    r &= u
    '''
    q = ', '.join(bab.q_vars)
    s = (
        '{u} /\ '
        '\A {q}:  ({v} /\ {p_leq_q}) => ({p_eq_q})').format(
            u=u,
            v=v,
            p_leq_q=bab.p_leq_q,
            p_eq_q=bab.p_eq_q,
            q=q)
    r = fol.add_expr(s)
    '''
    assert support_issubset(r, bab.p_vars, fol)
    return r


def _lower_bound(x, y, p_leq_q, p_to_q, fol):
    """Return lower bound (greedy)."""
    log.debug('---- lower bound ----')
    _, n = _independent_set(
        x, y, p_leq_q, p_to_q, fol, only_size=True)
    _assert_possible_cover_size(n, x, fol)
    log.info('lower bound = {n}'.format(n=n))
    log.debug('==== lower bound ====')
    return n


def _upper_bound(x, y, p_leq_q, p_to_q, fol):
    """Return upper bound (greedy)."""
    log.debug('---- upper bound ----')
    _, n = _some_cover(
        x, y, p_leq_q, p_to_q, fol, only_size=True)
    _assert_possible_cover_size(n, x, fol)
    log.info('upper bound = {n}'.format(n=n))
    log.debug('==== upper bound ====')
    return n


def _lower_bound_naive(x, y, p_leq_q, p_to_q, fol):
    """Return lower bound.

    Naive computation that greedily constructs
    an independent set.
    """
    log.debug('---- naive lower bound ----')
    z, n = _independent_set(x, y, p_leq_q, p_to_q, fol)
    assert support_issubset(z, p_to_q, fol)
    _assert_possible_cover_size(n, x, fol)
    log.info('lower bound = {n}'.format(n=n))
    log.debug('==== naive lower bound ====')
    return n


def _upper_bound_naive(x, y, p_leq_q, p_to_q, fol):
    """Return upper bound.

    Naive computation that greedily constructs
    an irredundant cover.
    """
    log.debug('---- naive upper bound ----')
    cover, n = _some_cover(x, y, p_leq_q, p_to_q, fol)
    assert support_issubset(cover, p_to_q, fol)
    _assert_possible_cover_size(n, x, fol)
    log.info('upper bound = {n}'.format(n=n))
    log.debug('==== naive upper bound ====')
    return n


def _assert_possible_cover_size(n, x, fol):
    """Raise `AssertionError` if size `u` not of a cover."""
    assert n >= 0, n
    assert (n == 0) is (x == fol.false), n


def _independent_set(
        x, y, p_leq_q, p_to_q, fol, only_size=False):
    """Return an independent set of implicants and its size.

    Requires that each implicant in `x` be covered
    by at least one implicant in `y`.

    @param only_size: Do not return the independent set.
        This avoids creating the BDD of a sparse set.
    """
    log.debug('---- independent set ----')
    p = set(p_to_q)
    q = set(p_to_q.values())
    assert support_issubset(x, p, fol), (fol.support(x), p)
    assert support_issubset(y, p, fol), (fol.support(y), p)
    yq = fol.let(p_to_q, y)
    assert _cover_refines(x, yq, p_leq_q, p, q, fol), r
    rem = x
    z = fol.false
    k = 0
    n = fol.count(rem)
    assert n >= 0, n
    while rem != fol.false:
        x0 = fol.pick(rem)
        assert set(x0) <= p, x0
        if not only_size:
            z |= fol.assign_from(x0)
        # umbrella of x0
        #
        # r(p) == \E q:  /\ x0 <= q
        #                /\ p <= q
        #                /\ y(q)
        r = yq & fol.let(x0, p_leq_q)
        r &= p_leq_q
        r = fol.exist(q, r)
        # update
        assert support_issubset(r, p, fol)
        rem &= ~ r
        k += 1
        # variant
        nold = n
        n = fol.count(rem)
        assert n < nold, (n, nold)
    assert fol.count(rem) == 0, 'some x not covered'
    _assert_possible_cover_size(k, x, fol)
    log.debug('==== independent set ====')
    if only_size:
        return None, k
    k_ = fol.count(z)
    assert k == k_, (k, k_)
    return z, k


def _some_cover(x, y, p_leq_q, p_to_q, fol, only_size=False):
    """Return a cover and its size, possibly not minimal.

    Signature similar to `_independent_set`.
    """
    log.debug('---- some cover ----')
    p = set(p_to_q)
    q = set(p_to_q.values())
    assert support_issubset(x, p, fol), (fol.support(x), p)
    assert support_issubset(y, p, fol), (fol.support(y), p)
    yq = fol.let(p_to_q, y)
    assert _cover_refines(x, yq, p_leq_q, p, q, fol), r
    rem = x
    z = fol.false
    k = 0
    n = fol.count(rem)
    assert n >= 0, n
    while rem != fol.false:
        x0 = fol.pick(rem)  # x0
        assert set(x0) <= p, x0
        # ys that cover x0
        #
        # r(q) == /\ x0 <= q
        #         /\ y(q)
        r = yq & fol.let(x0, p_leq_q)
        y0 = fol.pick(r)
        assert set(y0) <= q, y0
        if not only_size:
            z |= fol.assign_from(y0)
        # x that y0 does not cover
        # rem(p) /\ ~ (p <= y0)
        rem &= ~ fol.let(y0, p_leq_q)
        k += 1
        # variant
        nold = n
        n = fol.count(rem)
        assert n < nold, (n, nold)
    assert fol.count(rem) == 0, 'some x not covered'
    _assert_possible_cover_size(k, x, fol)
    log.debug('==== some cover ====')
    if only_size:
        return None, k
    q_to_p = {v: k for k, v in p_to_q.items()}
    zp = fol.let(q_to_p, z)
    k_ = fol.count(z)
    assert k == k_, (k, k_)
    return zp, k


def _cover_refines(xp, yq, p_leq_q, p, q, fol):
    """Return `True` if cover `xp` refines `yq`.

    In other words, if each implicant of `xp` is
    covered by some implicant in `yq` (let q == p).
    """
    # Each implicant in `x` must be covered
    # by some implicant in `y`.
    #
    # Return
    #     \A p:  \/ ~ x(p)
    #            \/ \E q:  /\ y(q)
    #                      /\ p <= q
    r = yq & p_leq_q
    r = fol.exist(q, r)
    r |= ~ xp
    r = fol.forall(p, r)
    return r == fol.true


def _partial_order(px, fol):
    """Return `u <= p` and `p <= u`."""
    ux = {
        x: dict(
            a=_prime_like(d['a']),
            b=_prime_like(d['b']))
        for x, d in px.items()}
    varmap = _parameter_varmap(ux, px)
    u_leq_p = _orthotope_subseteq(varmap, fol)
    varmap = _parameter_varmap(px, ux)
    p_leq_u = _orthotope_subseteq(varmap, fol)
    return u_leq_p, p_leq_u


def essential_orthotopes(f, px, qx, fol, xvars):
    """Return essential prime orthotopes of `f`."""
    log.info('---- essential orthotopes ----')
    varmap = _parameter_varmap(px, qx)
    p_leq_q = _orthotope_subseteq(varmap, fol)
    p_eq_q = _orthotope_eq(varmap, fol)
    p_is_prime = prime_orthotopes(
        f, px, qx,
        p_leq_q, p_eq_q,
        fol, xvars)
    # add to quantify u, v, so that we can rename
    #
    # TODO: reimplement using `fol.Context.rename`
    varmap = _parameter_varmap(qx, px)
    q_leq_p = _orthotope_subseteq(varmap, fol)
    q_is_prime = prime_orthotopes(
        f, qx, px,
        q_leq_p, p_eq_q,
        fol, xvars)
    x_in_q = _orthotope_contains_x(qx, fol)
    x = ', '.join(px)
    q = ', '.join(_collect_parameters(qx))
    s = (
        '{p_is_prime} /\ '
        r'\E {x}:  ( '
        '    {f} /\ '
        r'    \A {q}:  ( '
        '        ( '
        '            {q_is_prime} /\ '
        '            ~ {p_eq_q} '
        '        ) '
        '        => ~ {x_in_q}'
        '    )'
        ')').format(
            p_is_prime=p_is_prime,
            q_is_prime=q_is_prime,
            p_eq_q=p_eq_q,
            x_in_q=x_in_q,
            f=f, x=x, q=q)
    r = fol.add_expr(s)
    log.info('==== essential orthotopes ====')
    return r


def prime_orthotopes(
        f, px, qx,
        p_leq_q, p_eq_q,
        fol, xvars):
    """Return dominators of implicants."""
    log.info('----- prime orthotopes ----')
    assert support_issubset(f, xvars, fol)
    p_to_q = _renaming_between_parameters(px, qx)
    p_is_implicant = _implicant_orthotopes(f, px, fol, xvars)
    q_is_implicant = fol.let(p_to_q, p_is_implicant)
    q = _collect_parameters(qx)
    r = q_is_implicant & p_leq_q
    r = p_eq_q | ~ r
    r = fol.forall(q, r)
    r &= p_is_implicant
    '''
    q = ', '.join(_collect_parameters(qx))
    s = (
        '{p_is_implicant} /\ '
        r'\A {q}:  ( '
        '     ({q_is_implicant} /\ {p_leq_q})'
        '     => {p_eq_q}'
        ')').format(
            p_is_implicant=p_is_implicant,
            q_is_implicant=q_is_implicant,
            p_leq_q=p_leq_q,
            p_eq_q=p_eq_q,
            q=q)
    r = fol.add_expr(s)
    '''
    log.info('==== prime orthotopes ====')
    return r


def _implicant_orthotopes(f, abx, fol, xvars):
    """Return orthotopes that imply `f`.

    Caution: `fol` type hints are ignored.
    """
    log.info('---- implicant orthotopes ----')
    assert support_issubset(f, xvars, fol)
    x = ', '.join(abx)
    h = _orthotope_contains_x(abx, fol)
    nonempty = _orthotope_nonempty(abx, fol)
    s = (
        '{nonempty} /\ '
        '\A {x}:  {h} => {f} ').format(
            x=x, h=h, f=f, nonempty=nonempty)
    r = fol.add_expr(s)
    log.info('==== implicant orthotopes ====')
    return r


def _none_covered(
        cover_p, f,
        p_to_q, px, qx, fol):
    """Return `True` if `cover_p` covers no minterm in `f`.

    Arguments similar to `covers`.
    """
    p = set(p_to_q)
    q = set(p_to_q.values())
    varmap = _parameter_varmap(px, qx)
    fq = _embed_as_implicants(f, qx, fol)
    # \A p:  \/ ~ cover(p)
    #        \/ ~ \E q:  /\ f(q)
    #                    /\ Intersect(p, q)
    r = fq & _orthotopes_intersect(varmap, fol)
    r = ~ fol.exist(q, r)
    r |= ~ cover_p
    r = fol.forall(p, r)
    return r == fol.true


def _covers(
        cover_p, f, p_leq_q,
        p_to_q, px, fol):
    """Return `True` if `cover_p` covers `f`.

    @param cover_p: primes, repr as `p`
    @param f: elements to cover, repr as `x`
    @param px: mapping from `p` to intervals
    """
    p = set(p_to_q)
    q = set(p_to_q.values())
    fp = _embed_as_implicants(f, px, fol)
    cover_q = fol.let(p_to_q, cover_p)
    # \A p:  \/ ~ f(p)
    #        \/ \E q:  cover(q) /\ (p <= q)
    r = cover_q & p_leq_q
    r = fol.exist(q, r)
    r |= ~ fp
    r = fol.forall(p, r)
    return r == fol.true


def _covers_naive(cover_p, f, px, fol):
    """Return `True` if `cover_p` covers `f`.

    Same as `covers`. Here the computation happens over
    the concrete variables (`x`), so it is less efficient.
    """
    x_vars = set(px)
    assert support_issubset(f, x_vars, fol)
    # concretize
    x_in_cover = _concretize_implicants(cover_p, px, fol)
    covered = ~ f | x_in_cover
    return covered == fol.true


def _concretize_implicants(implicants_p, px, fol):
    """Return covered set as function of `x`."""
    # assert
    x_vars = set(px)
    p_vars = _collect_parameters(px)
    assert support_issubset(implicants_p, p_vars, fol)
    # concretize
    x_in_p = _orthotope_contains_x(px, fol)
    u = x_in_p & implicants_p
    covered_x = fol.exist(p_vars, u)
    assert support_issubset(covered_x, x_vars, fol)
    return covered_x


def _embed_as_implicants(f, px, fol):
    ax = {x: d['a'] for x, d in px.items()}
    bx = {x: d['b'] for x, d in px.items()}
    u = fol.let(ax, f)
    v = _orthotope_singleton(px, fol)
    return u & v


def _embed_as_implicants_naive(f, px, fol):
    """Return product representation of minterms."""
    x_as_ab = {x: dict(a=x, b=x) for x in px}
    varmap = _parameter_varmap(px, x_as_ab)
    r = _orthotope_eq(varmap, fol)
    return fol.exist(x_as_ab, r & f)


# slower than `_orthotope_singleton`
# needs `from omega.symbolic import fol as _fol`
# @profile
def _orthotope_singleton_wo_parsing(px, fol):
    """Return BDD that orthotope contains single point."""
    a_b = {d['a']: d['b'] for d in px.values()}
    bitmap = _fol._refine_renaming(a_b, fol.vars)
    bdd = fol.bdd
    r = bdd.true
    for b1, b2 in bitmap.items():
        u = bdd.var(b1)
        v = bdd.var(b2)
        u = bdd.apply('<=>', u, v)
        r &= u
    r_ = _orthotope_signleton(px, fol)
    assert r == r_, (r, r_)
    return r


def _orthotope_singleton(px, fol):
    """Return BDD that orthotope contains single point."""
    s = stx.conj(
        '({a} = {b})'.format(
            a=d['a'], b=d['b'])
        for x, d in px.items())
    r = fol.add_expr(s)
    return r


def _orthotope_nonempty(abx, fol):
    """Return condition that orthotope be non-empty."""
    s = stx.conj(
        '({a} <= {b})'.format(
            a=d['a'], b=d['b'])
        for x, d in abx.items())
    r = fol.add_expr(s)
    return r


def _orthotope_contains_x(abx, fol):
    """Return `x \in concretization(abx)`."""
    s = stx.conj((
        '({a} <= {x}) /\ '
        '({x} <= {b})').format(
            x=x, a=d['a'], b=d['b'])
        for x, d in abx.items())
    r = fol.add_expr(s)
    return r


def _orthotope_subseteq(varmap, fol):
    r"""Return `ab \subseteq uv`.

    This is the partial order defined by the subset relation.
    In the general formulation `\sqsubseteq`.
    """
    s = stx.conj((
        '({u} <= {a}) /\ '
        '({b} <= {v})').format(a=a, b=b, u=u, v=v)
            for (a, b), (u, v) in varmap.items())
    r = fol.add_expr(s)
    return r


def _orthotope_eq(varmap, fol):
    """Return `ab = uv`."""
    s = stx.conj((
        '({a} = {u}) /\ '
        '({b} = {v})').format(a=a, b=b, u=u, v=v)
        for (a, b), (u, v) in varmap.items())
    r = fol.add_expr(s)
    return r


def _orthotopes_intersect(varmap, fol):
    """Return `ab \cap uv # \emptyset`.

    Equivalent to

    \E x:  /\ x \in concretization(ab)
           /\ x \in concretization(uv)

    The representation of orthotopes as products of
    intervals allows for a direct construction that
    avoids quantification over `x`.
    """
    # disjoint intervals in at least one dimension
    s = stx.disj((
        '({b} < {u}) \/ '
        '({v} < {a})').format(a=a, b=b, u=u, v=v)
            for (a, b), (u, v) in varmap.items())
    r = fol.add_expr(s)
    return ~ r


def plot_orthotopes(u, abx, axvars, fol, ax):
    """Plot a polytope for each orthotope in `u`.

    @param axvars: `list` that defines which variable
        spans each dimension of the plot.
    """
    try:
        import polytope as poly
    except ImportError:
        raise ImportError(
            '`orthotopes` failed to import `polytope`.\n'
            'No plotting of orthotopes.')
    c = _orthotopes_iter(u, fol)
    eps = 0.1
    cycol = cycle('bgrcmk')
    for product in c:
        x, y = axvars
        a_x = abx[x]['a']
        b_x = abx[x]['b']
        a_y = abx[y]['a']
        b_y = abx[y]['b']
        xmin = product[a_x]
        xmax = product[b_x]
        ymin = product[a_y]
        ymax = product[b_y]
        # if a = b add a small amount
        if xmin == xmax:
            xmin -= eps
            xmax += eps
        if ymin == ymax:
            ymin -= eps
            ymax += eps
        size = [[xmin, xmax], [ymin, ymax]]
        p = poly.box2poly(size)
        color = next(cycol)
        p.plot(ax=ax, color=color, alpha=0.5)


def plot_points(u, keys, fol, ax, **kw):
    """Plot in 2D the integer assignments in a BDD."""
    for d in fol.pick_iter(u, care_vars=keys):
        plot(d, keys, ax=ax, **kw)


def plot(d, keys, ax, **kw):
    """The missing plot function for named coordinates."""
    t = tuple(d[k] for k in keys)
    ax.plot(*t, **kw)


def dumps_cover(
        cover, f, care, fol,
        latex=False,
        show_dom=False,
        show_limits=False):
    """Return disjunction of orthotopes in `cover`, one per line.

    @param latex: use `pf.sty` commands
    @param show_dom: if `care` implies type hints,
        then conjoin type hints (`fol.vars[var]['dom']`)
    @param show_limits: conjoin limits of  bitfield values

    @rtype: `str`
    """
    x_vars, px, _, _ = _setup_aux_vars(f, care, fol)
    c = list()
    if show_limits:
        r = _list_limits(x_vars, fol.vars)
        c.extend(r)
    show_dom = show_dom and _care_implies_type_hints(f, care, fol)
    if show_dom:
        r = _list_type_hints(x_vars, fol.vars)
        c.extend(r)
    else:
        log.info(
            'type hints omitted (care does not imply them)')
    r = _list_orthotope_expr(
        cover, px, fol, use_dom=show_dom)
    s = _vertical_op(r, op='or', latex=latex)
    c.append(s)
    n_expr = len(r)
    if care != fol.true:
        c.append('care expression')
    s = _vertical_op(c, op='and', latex=latex)
    # could add option to find minimal cover for care too
    return s


def _vertical_op(c, latex=False, op='and'):
    """Return TLA conjunction with one conjunct per line."""
    assert op in {'and', 'or'}, op
    if not c:
        r = 'TRUE' if op == 'and' else 'FALSE'
        return r
    # singleton ?
    if len(c) == 1:
        return c[0]
    if latex:
        pref = '  '
        nl = r' \\' + '\n'
    else:
        pref = '/\\' if op == 'and' else '\/'
        nl = '\n'
    r = list()
    for s in c:
        t = s.split('\n')
        t[0] = ' {p} {first}'.format(p=pref, first=t[0])
        e = '\n    '.join(t)
        r.append(e)
    r = nl.join(r)
    env = 'conj' if op == 'and' else 'disj'
    if latex:
        r = ('\\begin{' + env + '}\n' + r +
             '\n\\end{' + env + '}')
    return r


def _list_orthotope_expr(
        cover, px, fol,
        simple=False,
        use_dom=False):
    """Return `list` of `str`, each an orthotope in `cover`.

    @param simple: if `True`, then return expression
        that can be parsed by `fol.add_expr`.
    @param use_dom: omit conjuncts that contain dom of var
        assumes that `|= care => type_hints`
    """
    xp = _map_parameters_to_vars(px)
    support = fol.support(cover)
    keys = {xp[k] for k in support}
    keys = natsort.natsorted(keys)
    c = _orthotopes_iter(cover, fol)
    r = list()
    for product in c:
        w = list()
        for x in keys:
            a = px[x]['a']
            b = px[x]['b']
            a = product[a]
            b = product[b]
            _check_type_hint(a, b, fol.vars[x], x)
            # can `x` be ignored ?
            if use_dom:
                dom = fol.vars[x]['dom']
                a, b = _clip_subrange((a, b), dom, x)
            if a is None and b is None:
                continue
            if a == b:
                s = '({x} = {a})'
            elif simple:
                s = '({a} <= {x}) /\ ({x} <= {b})'
            else:
                # precise even in absence of limits/dom
                s = '({x} \in {a} .. {b})'
            s = s.format(x=x, a=a, b=b)
            w.append(s)
        s = ' /\ '.join(w)
        if not s:
            s = 'TRUE'
        r.append(s)
    r = natsort.natsorted(r)  # reproducable vertical listing
    return r


def _clip_subrange(ab, dom, x):
    """Return `ab` clipped to `dom`."""
    a, b = ab
    u, v = dom
    # assert not disjoint ranges
    assert a <= v and b >= u, (ab, dom, x)
    a = max(a, u)
    b = min(b, v)
    if a == u and v == b:
        a, b = None, None
    return a, b


def _check_type_hint(a, b, hint, var):
    """Raise `AssertionError` and log warnings."""
    if a > b:
        raise AssertionError((
            'Empty orthotope interval `{a} .. {b}` for '
            'variable "{var}".').format(
                a=a, b=b, var=var))
    dom = hint['dom']
    limits = _bitfield_limits(hint)
    if a > dom[1] or b < dom[0]:
        log.warning((
            'Interval `a .. b = {a} .. {b}` is disjoint from '
            'type hint `dom = {dom[0]} .. {dom[1]}` for '
            'variable "{var}".\n'
            'Use type hint `dom` as care set.\n').format(
                a=a, b=b, dom=dom, var=var))
    # `a <= dom[0]` should imply `a = limits[0]`,
    # because otherwise only possible if some point
    # outside type hint matters, preventing prime to
    # extend to the limits of the bitfield
    if a <= dom[0] and a > limits[0]:
        log.warning((
            'a = {a} not in interior of type hint '
            '`dom = {dom[0]} .. {dom[1]}` but unequal to '
            'low of `limits = {limits[0]} .. {limits[1]}` '
            'for variable "{var}".\n'
            'Use type hint as care set.\n').format(
                a=a, dom=dom, limits=limits, var=var))
    if b >= dom[1] and b < limits[1]:
        log.warning((
            'b = {b} not in interior of type hint '
            '`dom = {dom[0]} .. {dom[1]}` but unequal to '
            'high of `limits = {limits[0]} .. {limits[1]}` '
            'for variable "{var}".\n'
            'Use type hint as care set.\n').format(
                b=b, dom=dom, limits=limits, var=var))


def _care_implies_type_hints(f, care, fol):
    """Return `True` if `|= care => type_hints`."""
    vrs = joint_support([f, care], fol)
    type_hints = _conjoin_type_hints(vrs, fol)
    u = type_hints | ~ care
    return u == fol.true


def _f_implies_care(f, care, fol):
    """Return `True` if `|= f => care`."""
    return (care | ~ f) == fol.true


def _list_type_hints(variables, table):
    """Return `list` of set containment constraints.

    These are the high-level type invariant,
    as defined by the type hints (`"dom"`) in `table`.
    """
    assert variables, variables
    r = list()
    keys = natsort.natsorted(variables)
    for x in keys:
        dom = table[x]['dom']
        s = _format_range(x, *dom)
        r.append(s)
    return r


def _list_limits(vrs, table):
    """Return `list` of bitfield limits.

    Each limit has the form `x \in a .. b`

    @param table: must be bitblasted
    """
    assert vrs, vrs
    r = list()
    keys = natsort.natsorted(vrs)
    for x in keys:
        limits = _bitfield_limits(table[x])
        s = _format_range(x, *limits)
        r.append(s)
    return r


def _bitfield_limits(hint):
    """Return extremal integer values of bitfield."""
    width = hint['width']
    if hint['signed']:
        n = width - 1
        limits = (- 2**n, 2**n - 1)
        return limits
    # unsigned: so variable ranges over values of same sign
    min_, max_ = hint['dom']
    # flip ?
    if min_ >= 0:
        limits = (0, 2**width - 1)
    else:
        assert max_ < 0, max_
        limits = (- 2**width, -1)
    return limits


def _conjoin_type_hints(vrs, fol):
    """Return conjunction of type hints for `vrs` as BDD."""
    r = list()
    for var in vrs:
        hints = fol.vars[var]
        a, b = hints['dom']
        s = '({a} <= {var}) /\ ({var} <= {b})'.format(
            a=a, b=b, var=var)
        r.append(s)
    u = fol.add_expr(stx.conj(r))
    return u


def _format_range(var, a, b):
    """Return string for set containment `var \in a .. b`."""
    return r'{var} \in {a} .. {b}'.format(
        var=var, a=a, b=b)


def _orthotopes_iter(u, fol):
    """Return iterator over orthotopes."""
    if u == fol.false:
        log.info('empty set')
    c = fol.pick_iter(u)
    return c


def _setup_aux_vars(f, care, fol):
    """Add and return auxiliary variables.

    No BDD operations other than `support` are invoked.

    Returns:

    - `x_vars`: set of variable names in
        `support(f) \/ support(care)`
    - `px`: map var name to `dict` of indexed parameters
    - `qx`: similar for var copies
    - `p_to_q`: `dict` that maps parameters to their copies

    For example:

    ```
    x_vars = {'x', 'y'}
    px = dict(
        x=dict(a='a_x', b='b_x'),
        y=dict(a='a_y', b='b_y'))
    qx = dict(
        x=dict(a='u_x', b='v_x'),
        y=dict(a='u_y', b='v_y'))
    p_to_q = dict(
        a_x='u_x', b_x='v_x',
        a_y='u_y', b_y='v_y')
    ```

    @return x_vars, px, qx, p_to_q
    """
    assert f != fol.false
    assert care != fol.false
    assert not (f == fol.true and care == fol.true)
    x_vars = joint_support([f, care], fol)
    assert x_vars, x_vars
    # aux vars for orthotope representation
    params = dict(pa='a', pb='b', qa='u', qb='v')
    p_dom = _parameter_table(
        x_vars, fol.vars, a=params['pa'], b=params['pb'])
    q_dom = _parameter_table(
        x_vars, fol.vars, a=params['qa'], b=params['qb'])
    p_dom = _add_prime_like_too(p_dom)
    q_dom = _add_prime_like_too(q_dom)
    common = x_vars.intersection(p_dom)
    assert not common, common
    common = x_vars.intersection(q_dom)
    assert not common, common
    # works for primed variables too
    fol.declare(**p_dom)
    fol.declare(**q_dom)
    px = _parameter_variables(x_vars, a=params['pa'], b=params['pb'])
    qx = _parameter_variables(x_vars, a=params['qa'], b=params['qb'])
    assert set(px) == set(qx), (px, qx)
    p_to_q = _renaming_between_parameters(px, qx)
    log.debug('x vars: {x_vars}'.format(x_vars=x_vars))
    return x_vars, px, qx, p_to_q


def _parameter_table(x, table, a, b):
    """Return symbol table that defines parameters.

    Supports integer-valued variables only.
    Represent Boolean-valued as 0..1-valued variables.
    """
    assert x, x
    d = dict()
    for xj in x:
        dtype = table[xj]['type']
        assert dtype in ('int', 'saturating'), dtype
        dom = table[xj]['dom']
        name = _replace_prime(xj)
        aj = '{a}_{v}'.format(a=a, v=name)
        bj = '{b}_{v}'.format(b=b, v=name)
        d[aj] = dom
        d[bj] = dom
        assert "'" not in aj, aj
        assert "'" not in bj, bj
    assert len(d) == 2 * len(x), d
    return d


def _parameter_variables(x_vars, a, b):
    """Return `dict` that maps each var x to a_x, b_x."""
    d = dict()
    for x in x_vars:
        name = _replace_prime(x)
        a_x = '{a}_{v}'.format(v=name, a=a)
        b_x = '{b}_{v}'.format(v=name, b=b)
        d[x] = dict(a=a_x, b=b_x)
    return d


def _map_parameters_to_vars(px):
    """Return map `{a: x, b: x, ...}`."""
    d = {d['a']: k for k, d in px.items()}
    d.update((d['b'], k) for k, d in px.items())
    return d


def _collect_parameters(px):
    """Return `set` of parameters from `px`."""
    c = set()
    c.update(d['a'] for d in px.values())
    c.update(d['b'] for d in px.values())
    assert len(c) == 2 * len(px), (c, px)
    return c


def _parameter_varmap(px, qx):
    """Return map `{(a, b): (u, v), ... }`."""
    assert set(px) == set(qx), (px, qx)
    d = dict()
    for x in px:
        a = px[x]['a']
        b = px[x]['b']
        u = qx[x]['a']
        v = qx[x]['b']
        d[(a, b)] = (u, v)
    return d


def _renaming_between_parameters(px, qx):
    """Return map `{a: u, b: v, ... }`."""
    assert set(px) == set(qx), (px, qx)
    d = dict()
    for x in px:
        a = px[x]['a']
        b = px[x]['b']
        u = qx[x]['a']
        v = qx[x]['b']
        d[a] = u
        d[b] = v
    return d


def _replace_prime(var):
    """Replace postfix prime with "_p"

    To avoid parsing a parameter's name as if it is
    two names with an operator in the middle.
    This case arises for parameters that associated to
    primed variables.

    For example, when computing a minimal cover for
    an action.
    """
    if not stx.isprimed(var):
        assert "'" not in var, var
        return var
    assert stx.isprimed(var), var
    unprimed = stx.unprime(var)
    # `'` even in the middle would split it when parsing
    assert "'" not in unprimed, unprimed
    var_p = '{unprimed}_p'.format(unprimed=unprimed)
    assert "'" not in var_p, var_p
    return var_p


def _add_prime_like_too(table):
    """Return new table of primed and unprimed vars.

    All variables in `table` should be unprimed.

    @type table: `dict`
    @rtype: `dict`
    """
    t = dict()
    for var, dom in table.items():
        assert not stx.isprimed(var), var
        pvar = _prime_like(var)
        if dom == 'bool':
            r = 'bool'
        else:
            assert len(dom) == 2, dom
            r = tuple(dom)
        t[var] = r
        t[pvar] = r
    return t


def _prime_like(var):
    return '{var}_cp'.format(var=var)


class _BranchAndBound(object):
    """A data structure that stores useful values.

    It helps avoid passing 10 arguments in each
    function call during the recursion.
    Attributes:

    - `lower_bound`: global lower bound
    - `upper_bound`: global upper bound (feasible)
    - `p_leq_q`: partial order `p <= q`
        (anti-symmetry required: a quasi-order does not work.)
    - `p_to_q`: mapping from `p` to `q`
    - `u_leq_p`: partial order `u <= p`
    """

    def __init__(
            self, p_leq_q, p_leq_u, u_leq_p,
            p_eq_q, p_to_q, px, qx, fol):
        p_vars = set(p_to_q)
        q_vars = set(p_to_q.values())
        p_to_u = {p: _prime_like(p) for p in p_vars}
        u_vars = set(p_to_u.values())
        assert not (p_vars & q_vars), (p_vars, q_vars)
        assert not (p_vars & u_vars), (p_vars, u_vars)
        assert not (u_vars & q_vars), (u_vars, q_vars)
        self.p_vars = p_vars
        self.q_vars = q_vars
        self.u_vars = u_vars
        pq_vars = p_vars.union(self.q_vars)
        assert support_issubset(p_leq_q, pq_vars, fol)
        self._lower_bound = None
        self._upper_bound = None
        self.best_cover = None  # found so far
        self.p_leq_q = p_leq_q
        self.p_leq_u = p_leq_u
        self.u_leq_p = u_leq_p
        self.p_eq_q = p_eq_q
        self.p_to_q = p_to_q
        self.p_to_u = p_to_u
        self.px = px
        self.qx = qx
        self.fol = fol

    def _assert_invariant(self, lower=None, upper=None):
        """Raise `AssertionError` if lower > upper bound."""
        if lower is None:
            lower = self._lower_bound
        if upper is None:
            upper = self._upper_bound
        if lower is None or upper is None:
            return
        assert lower <= upper, (lower, upper)

    @property
    def lower_bound(self):
        return self._lower_bound

    @lower_bound.setter
    def lower_bound(self, c):
        assert c >= 0, c
        # only initialize global lower bound
        assert self._lower_bound is None, self._lower_bound
        self._assert_invariant(lower=c)
        self._lower_bound = c

    @property
    def upper_bound(self):
        return self._upper_bound

    @upper_bound.setter
    def upper_bound(self, c):
        assert c > 0, c
        self._assert_invariant(upper=c)
        if self._upper_bound is None:
            log.info(
                'initialized upper bound to {c}'.format(
                    c=c))
        elif c < self._upper_bound:
            log.info((
                'improved upper bound from '
                '{old} to {new}').format(
                    old=self._upper_bound,
                    new=c))
        self._upper_bound = c



"""Convert relation predicate to predicate-action diagram.

Reference
=========

Leslie Lamport
    "TLA in pictures"
    IEEE Transactions on Software Engineering
    Vol.21, No.9, pp.768--775, 1995
    doi: 10.1109/32.464544
"""
# Copyright 2016 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import collections
from itertools import chain
import logging

import networkx as nx

from omega.logic import syntax as stx
from omega.symbolic import fol as _fol
from omega.symbolic import symbolic


log = logging.getLogger(__name__)


def action_to_steps(aut, qinit='\A \A'):
    r"""Return enumerated graph with steps as edges.

    Only `aut.init['env']` considered.
    The predicate `aut.init['sys']` is ignored.

    `qinit` has different meaning that in `omega.games.gr1`.
    Nonetheless, for synthesized `aut.init['env']`,
    the meaning of `qinit` here yields the expected result.

    Enumeration is done based on `qinit`:

    - `'\A \A'`: pick all states that satisfy `aut.init['env']`
    - `'\E \E'`: pick one state that satisfies `aut.init['env']`
    - `'\A \E'`: for all states that satisfy `aut.init['env']`,
      pick a unique state for each env state `x`
    - `'\E \A'`: pick a sys state `u` and enumerate all
      states that satisfy `aut.init['env']` and `y = u`
    """
    assert aut.action['sys'] != aut.false
    primed_vars = _primed_vars_per_quantifier(aut.varlist)
    unprime_vars = {stx.prime(var): var for var in aut.vars
                    if not stx.isprimed(var)}
    # fix an order for tupling
    keys = list(k for k in aut.vars if not stx.isprimed(k))
    umap = dict()  # map assignments -> node numbers
    g = nx.DiGraph()
    queue, visited = _init_search(g, aut, umap, keys, qinit)
    g.initial_nodes = set(queue)
    varnames = set(keys)
    symbolic._assert_support_moore(aut.action['sys'], aut)
    # search
    while queue:
        node = queue.pop()
        values = g.node[node]
        log.debug('at node: {d}'.format(d=values))
        assert set(values) == varnames, (values, aut.vars)
        u = aut.action['env']
        u = aut.let(values, u)
        # apply Mealy controller function
        env_iter = aut.pick_iter(
            u, care_vars=primed_vars['env'])
        u = aut.action['sys']
        assert u != aut.false
        sys = aut.let(values, u)
        assert sys != aut.false
        for next_env in env_iter:
            log.debug('next_env: {r}'.format(r=next_env))
            # no effect if `aut.moore`
            u = aut.let(next_env, sys)
            u = aut.let(unprime_vars, u)
            env_values = {unprime_vars[var]: value
                          for var, value in next_env.items()}
            v = aut.let(env_values, visited)
            # prefer already visited nodes
            v &= u
            if v == aut.false:
                log.info('cannot remain in visited nodes')
                v = u
                remain = False
            else:
                remain = True
            assert v != aut.false
            sys_values = aut.pick(
                v, care_vars=aut.varlist['sys'])
            d = dict(env_values)
            d.update(sys_values)
            # assert
            u = aut.let(d, visited)
            assert u == aut.true or u == aut.false
            assert remain == (u == aut.true), remain
            # find or add node
            if remain:
                next_node = _find_node(d, umap, keys)
            else:
                next_node = _add_new_node(d, g, queue, umap, keys)
                visited = _add_to_visited(d, visited, aut)
            g.add_edge(node, next_node)
            log.debug((
                'next env: {e}\n'
                'next sys: {s}\n').format(
                    e=env_values,
                    s=sys_values))
    return g


def _primed_vars_per_quantifier(varlist):
    """Return `dict` that maps each player to set of primed vars.

    @return: `dict` that maps
          player name -> set of primed vars
    """
    assert 'env' in varlist, varlist
    assert 'sys' in varlist, varlist
    primed_vars = dict(
        env={stx.prime(var) for var in varlist['env']},
        sys={stx.prime(var) for var in varlist['sys']})
    return primed_vars


def _init_search(g, aut, umap, keys, qinit):
    """Enumerate initial states according to `qinit`."""
    # danger of blowup due to sparsity
    # implement enumerated equivalent to compare
    if qinit == '\A \E':
        queue, visited = _forall_exist_init(g, aut, umap, keys)
    elif qinit == '\A \A':
        queue, visited = _forall_init(g, aut, umap, keys)
    elif qinit == '\E \E':
        queue, visited = _exist_init(g, aut, umap, keys)
    elif qinit == '\E \A':
        queue, visited = _exist_forall_init(g, aut, umap, keys)
    else:
        raise Exception('unknown qinit "{q}"'.format(q=qinit))
    log.info('{n} initial nodes'.format(n=len(queue)))
    return queue, visited


def _forall_init(g, aut, umap, keys):
    r"""Enumerate initial states with \A \A vars."""
    env_init = aut.init['env']
    assert env_init != aut.false
    init_iter = aut.pick_iter(
        env_init,
        care_vars=chain(aut.varlist['env'], aut.varlist['sys']))
    visited = aut.false
    queue = list()
    for d in init_iter:
        _add_new_node(d, g, queue, umap, keys)
        visited = _add_to_visited(d, visited, aut)
    return queue, visited


def _exist_init(g, aut, umap, keys):
    r"""Enumerate initial states with \E env, sys vars."""
    env_init = aut.init['env']
    assert env_init != aut.false
    d = aut.pick(
        env_init,
        care_vars=chain(aut.varlist['env'], aut.varlist['sys']))
    visited = aut.false
    queue = list()
    _add_new_node(d, g, queue, umap, keys)
    visited = _add_to_visited(d, visited, aut)
    return queue, visited


def _forall_exist_init(g, aut, umap, keys):
    r"""Enumerate initial states with \A env:  \E sys vars.

    Note that each initial "state" is a class of
    initial states in ZF set theory.
    """
    env_init = aut.init['env']
    assert env_init != aut.false
    only_env_init = aut.exist(aut.varlist['sys'], env_init)
    env_iter = aut.pick_iter(
        only_env_init, care_vars=aut.varlist['env'])
    visited = aut.false
    queue = list()
    for env_0 in env_iter:
        u = aut.let(env_0, env_init)
        sys_0 = aut.pick(u, care_vars=aut.varlist['sys'])
        d = dict(env_0)
        d.update(sys_0)
        # confirm `sys_0` picked properly
        u = aut.let(d, env_init)
        assert u == aut.true, u
        _add_new_node(d, g, queue, umap, keys)
        visited = _add_to_visited(d, visited, aut)
    return queue, visited


def _exist_forall_init(g, aut, umap, keys):
    r"""Enumerate initial states with \E sys:  \A env vars."""
    # this function can be merged with `_forall_exist_init`
    # by constraining initial sys assignments,
    # then enumerating the same way
    env_init = aut.init['env']
    assert env_init != aut.false
    # pick `sys_0` so that it work for all
    # env assignments alowed by `env_init`
    only_env_init = aut.exist(aut.varlist['sys'], env_init)
    u = ~ only_env_init | env_init
    u = aut.forall(aut.varlist['env'], u)
    assert u != aut.false
    sys_0 = aut.pick(u, care_vars=aut.varlist['sys'])
    # iterate over env initial assignments
    # independently of sys
    env_iter = aut.pick_iter(
        only_env_init, care_vars=aut.varlist['env'])
    visited = aut.false
    queue = list()
    for env_0 in env_iter:
        d = dict(env_0)
        d.update(sys_0)
        # confirm `sys_0` works for all `env_0`
        u = aut.let(d, env_init)
        assert u == aut.true, u
        _add_new_node(d, g, queue, umap, keys)
        visited = _add_to_visited(d, visited, aut)
    return queue, visited


def _find_node(d, umap, keys):
    """Return node in `umap` with assignment `d`."""
    assert isinstance(keys, collections.Sequence), keys
    key = _node_tuple(d, keys)
    assert key in umap, (key, umap)
    u = umap[key]
    return u


def _add_new_node(d, g, queue, umap, keys):
    """Add node to graph `g` for the assignment `d`."""
    assert isinstance(keys, collections.Sequence), keys
    u = len(g)
    assert u not in g, u
    g.add_node(u, **d)
    key = _node_tuple(d, keys)
    assert key not in umap, (key, umap)
    umap[key] = u
    queue.append(u)
    log.debug(d)
    return u


def _add_to_visited(values, visited, aut):
    """Return BDD `visted` updated with assignment `values`."""
    c = list()
    for var, value in values.items():
        t = aut.vars[var]['type']
        if t == 'bool':
            assert value in (True, False), value
            if bool(value):
                c.append(var)
            else:
                c.append(' ~ ' + var)
            continue
        # integer
        assert t in ('int', 'saturating', 'modwrap'), t
        s = '{var} = {value}'.format(var=var, value=value)
        c.append(s)
    s = stx.conj(c)
    u = aut.add_expr(s)
    visited |= u
    return visited


def _node_tuple(d, keys):
    """Return `tuple` of `d` values ordered by `keys`.

    @type d: `dict`
    @type keys: `tuple`
    """
    return tuple(d[k] for k in keys)



"""Symbolic automaton.


DEPRECATED: use `omega.symbolic.temporal` instead.


Including bitblasting of first-order to bitvector logic,
and transformation to BDD nodes.
"""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import copy
import logging
import pprint
import warnings

import natsort

from omega.logic import bitvector as bv
from omega.logic import syntax as stx
from omega.symbolic import bdd as sym_bdd


logger = logging.getLogger(__name__)
try:
    from dd import cudd as _bdd
    logger.info('`omega.symbolic.symbolic` will use `dd.cudd`')
except ImportError:
    from dd import autoref as _bdd
    logger.warning(
        '`omega.symbolic.symbolic` failed to import `dd.cudd`.\n'
        'Will use `dd.autoref`.')


class Automaton(object):
    r"""Transition relation, initial condition, and hidden variables.

    User-defined attributes
    =======================

    You define the attributes:

      - vars: `dict` that maps each variable name to
        a `dict` with keys:

          - `"type": "bool" or "modwrap" or "saturating"`
          - `"dom": (min, max)`, range of integer
          - `"owner": "env" or "sys"`
          - `"init"`: has suitable `__str__` (and is optional)

      - init: initial condition
      - action: transition relation
      - win: winning condition
      - acceptance: `str` that describes `win`,
          for example `'Streett(1)'`
      - moore: choose between Moore or Mealy implementation
      - plus_one: strict implication with `\weakprevious`
      - qinit: quantification of initial variable values:

        - `'\A \A'`
        - `'\A \E'`
        - `'\E \A'`
        - `'\E \E'`

    Each of `init, action` is a `dict`:

      `{player: list(), ...}`

    `win` is a `dict` of the form:

      `{'<>[]': list(),
        '[]<>': list()}`

    Add formulas (as strings) to these lists.
    Call the method `build` to convert these formulas to BDD nodes,
    and generate the below.


    Auto-generated attributes
    =========================

    The method `build` adds to `vars[var]` the keys:

      - `"bitnames"`: `list`
      - `"signed"`: `True` if signed integer
      - `"width"`: `len(bitnames)`

    Each of the following is a `set` of variable names:

      - uvars
      - upvars
      - ubvars
      - evars
      - epvars
      - ebvars
      - uevars
      - uepvars
      - prime: `dict` that maps each unprimed to a primed var
      - unprime: `dict` that maps eacn primed to an unprimed var

    Prefix meaning:

      u = universally quantified
      e = existentially quantified
      p = primed (absence means unprimed)
      b = both primed and unprimed


    References
    ==========

    Leslie Lamport
        "The temporal logic of actions"
        ACM Transactions on Programming
        Languages and Systems (TOPLAS),
        Vol.16, No.3, pp.872--923, 1994

    Rajeev Alur, Thomas A. Henzinger, Orna Kupferman
        "Alternating-time temporal logic"
        Journal of the ACM (JACM)
        Vol.49, No.5, pp.672--713, 2002

    Klaus Schneider
        "Verification of reactive systems"
        Springer, 2004
    """

    def __init__(self):
        warnings.warn(
            'Deprecated; use the class '
            '`omega.symbolic.temporal.Automaton` instead.',
            category=PendingDeprecationWarning)
        self.players = dict(env=0, sys=1)  # name -> turn
        self.vars = dict()
        self.moore = True
        self.plus_one = True
        self.qinit = '\A \A'
        # formulas
        self.init = dict(env=list(), sys=list())
        self.action = dict(env=list(), sys=list())
        self.win = {'<>[]': list(), '[]<>': list()}
        self.acceptance = 'Streett(1)'
        # auto-populated
        self.turns = list()  # inverse of `self.players`
        # subsets of bits
        self.uvars = set()  # unprimed env
        self.upvars = set()  # primed env
        self.ubvars = set()  # both primed and unprimed env
        self.evars = set()  # unprimed sys
        self.epvars = set()  # primed sys
        self.ebvars = set()  # both primed and unprimed env
        self.uevars = set()  # all unprimed vars
        self.uepvars = set()  # all primed vars
        # future: hidden, node vars
        # map between primed and unprimed
        self.prime = dict()  # unprimed -> primed
        self.unprime = dict()  # primed -> unprimed
        # aux
        # init only to aid static analysis
        self.bdd = _bdd.BDD()

    def __copy__(self):
        a = Automaton()
        a.players = copy.deepcopy(self.players)
        a.vars = copy.deepcopy(self.vars)
        a.init = copy.deepcopy(self.init)
        a.action = copy.deepcopy(self.action)
        a.win = copy.deepcopy(self.win)
        a.acceptance = self.acceptance
        a.bdd = self.bdd
        return a

    def __str__(self):
        c = [
            'Symbolic automaton:',
            40 * '=',
            '',
            'variables:',
            '{dvars}',
            '',
            'Moore' if self.moore else 'Mealy',
            ('causal' if self.plus_one else
             'circular') + ' implication',
            '',
            'Initially:',
            self.qinit]
        if self.init['env']:
            c.extend([
                'ENV INIT:',
                '{env_init}',
                ''])
        if self.action['env']:
            c.extend([
                'ENV ACTION:',
                '{env_action}',
                ''])
        if self.init['sys']:
            c.extend([
                'SYS INIT:',
                '{sys_init}',
                ''])
        if self.action['sys']:
            c.extend([
                'SYS ACTION:',
                '{sys_action}',
                ''])
        if self.win['<>[]']:
            c.extend([
                '<>[] (persistence):',
                '{fg}',
                ''])
        if self.win['[]<>']:
            c.extend([
                '[]<> (recurrence):',
                '{gf}',
                ''])
        return '\n'.join(c).format(
            dvars=pprint.pformat(self.vars),
            env_init=_join(self.init['env']),
            env_action=_join(self.action['env']),
            fg=_join(self.win['<>[]']),
            sys_init=_join(self.init['sys']),
            sys_action=_join(self.action['sys']),
            gf=_join(self.win['[]<>']))

    def dumps(self):
        """Return `repr` of a `dict` with the attributes."""
        d = dict(
            vars=self.vars,
            init=self.init,
            action=self.action,
            win=self.win,
            acceptance=self.acceptance,
            moore=self.moore,
            plus_one=self.plus_one,
            qinit=self.qinit)
        return repr(d)

    def build(self):
        """Return `Automaton` with formulas as BDD nodes.

        Bitblast variables and formulas,
        add bits to `self.bdd`,
        and populate primed and unprimed bits.

        BDD levels defined only if no existing bits.
        Otherwise, reorder to desired order later.
        """
        aut = _bitblast(self)
        aut.bdd = self.bdd
        aut = _bitvector_to_bdd(aut)
        aut.moore = self.moore
        aut.plus_one = self.plus_one
        aut.qinit = self.qinit
        return aut

    def assert_consistent(self, built=False):
        """Raise `AssertionError` if not well-formed.

        @param built: if `True`,
        then assert `build` has been called.
        """
        # check attributes
        for d in (self.init, self.action, self.win):
            for k, v in d.items():
                for u in v:
                    if built:
                        try:
                            assert u in self.bdd, u
                        except TypeError:
                            raise AssertionError(
                                'Automaton needs '
                                'to be compiled')
                    else:
                        assert isinstance(u, str)
        if not built:
            return
        # bdd initialized ?
        assert self.bdd.vars is not None
        if self.vars:
            assert self.uevars

    def update(self, attr, d):
        """Add formulas from `dict` `d`."""
        r = getattr(self, attr)
        if attr in ('init', 'action'):
            r['env'].extend(d['env'])
            r['sys'].extend(d['sys'])
        elif attr == 'win':
            r['<>[]'].extend(d['<>[]'])
            r['[]<>'].extend(d['[]<>'])
        else:
            raise Exception('unknown attr "{a}"'.format(a=attr))

    def conjoin(self, as_what):
        """Conjoin attributes.

        @param as_what: `'bdd' or 'prefix' or 'infix'`
        """
        _conj_owner(self, 'env', as_what)
        _conj_owner(self, 'sys', as_what)

    def add_expr(self, e):
        """Add first-order formula."""
        # the bitblaster understands priming
        # the prefix parser treats primed bits as fresh
        # so suffice it for t to contain unprimed only
        t = self.vars
        s = bv.bitblast(e, t)
        u = sym_bdd.add_expr(s, self.bdd)
        return u


def fill_blanks(aut, as_bdd=False, rabin=False):
    """Add `"True"` to empty attributes `init`, `action`, `win`.

    @param as_bdd: if `True`, then represent `"True"` as `1`
    """
    if as_bdd:
        true = aut.bdd.true
        false = aut.bdd.false
    else:
        true = 'True'
        false = 'False'
    for d in (aut.init, aut.action):
        for k, v in d.items():
            if not v:
                d[k] = [true]
    if not aut.win['<>[]']:
        if rabin:
            aut.win['<>[]'] = [true]
        else:
            aut.win['<>[]'] = [false]
    if not aut.win['[]<>']:
        aut.win['[]<>'] = [true]
    # post-condition
    for d in (aut.init, aut.action, aut.win):
        for k in d:
            assert len(d[k]) > 0


def _bitblast(aut):
    """Return `Automaton` with bitvector formulas.

    For each integer, the corresponding list of
    bitnames is added to the symbol table (attr `vars`).

    @type aut: `Automaton`
    """
    players = dict(aut.players)
    aut = copy.copy(aut)
    t = bv.bitblast_table(aut.vars)
    init, action = bv.type_invariants(t)
    for var, c in init.items():
        owner = aut.vars[var]['owner']
        # collect type invariants of parameters too,
        # for convenience later
        aut.init.setdefault(owner, list())
        aut.init[owner].extend(c)
    for var, c in action.items():
        owner = aut.vars[var]['owner']
        aut.action.setdefault(owner, list())
        aut.action[owner].extend(c)
    # conjoin to avoid it later over BDD nodes
    _conj_owner(aut, 'env', 'infix')
    _conj_owner(aut, 'sys', 'infix')
    a = Automaton()
    a.players = players
    a.vars = t
    # TODO: replace GR(1) syntax check
    # spec.check_syntax()
    _bitblast_owner(aut, a, 'env', t)
    _bitblast_owner(aut, a, 'sys', t)
    return a


def _conj_owner(aut, owner, as_what):
    """Conjoin the lists in the attributes of `owner`.

    @param as_what: `'bdd' or 'prefix' or 'infix'`
    """
    # get
    init = aut.init[owner]
    action = aut.action[owner]
    # compute
    if as_what == 'prefix':
        init = stx.conj_prefix(init)
        action = stx.conj_prefix(action)
    elif as_what == 'infix':
        init = stx.conj(init)
        action = stx.conj(action)
    elif as_what == 'bdd':
        def f(x, y):
            return aut.bdd.apply('and', x, y)
        init = stx.recurse_binary(f, init)
        action = stx.recurse_binary(f, action)
    else:
        raise Exception(
            'unknown as_what="{s}"'.format(s=as_what))
    # set
    aut.init[owner] = [init]
    aut.action[owner] = [action]


def _bitblast_owner(aut, a, owner, t):
    """Bitblast init, action, win of `owner`."""
    def f(x):
        return bv.bitblast(x, t)
    assert owner in ('env', 'sys'), owner
    a.init[owner] = list(map(f, aut.init[owner]))
    a.action[owner] = list(map(f, aut.action[owner]))
    if owner == 'env':
        s = '<>[]'
    else:
        s = '[]<>'
    a.win[s] = list(map(f, aut.win[s]))


def _bitvector_to_bdd(aut):
    """Return `Automaton` with BDD formulas.

    @type aut: `Automaton`
    """
    players = dict(aut.players)
    table = aut.vars
    bits = bv.bit_table(table, table)
    # index both, to allow for unquantified parameters
    ubits = set(b for b, d in bits.items()
                if d['owner'] == 'env')
    ebits = set(b for b, d in bits.items()
                if d['owner'] == 'sys')
    b = _pick_var_order(bits, ubits)
    b = _add_primed_bits(b)
    bdd = aut.bdd
    # define levels only if no existing vars
    if bdd.vars:
        for var in b:
            bdd.add_var(var)
    else:
        for level, var in enumerate(b):
            bdd.add_var(var, level)
    # bundle as:
    a = Automaton()
    a.vars = copy.deepcopy(table)
    a.players = players
    a.bdd = bdd
    # slugsin -> BDD
    # add long attributes first,
    # to improve  effectiveness of reordering.
    # better if each attr is one formula.
    from_sections = _make_section_map(aut)
    to_sections = _make_section_map(a)
    lengths = {
        k: _section_len(v)
        for k, v in from_sections.items()}
    sort = sorted(
        from_sections,
        key=lengths.__getitem__,
        reverse=True)
    for section in sort:
        p = from_sections[section]
        q = to_sections[section]
        _to_bdd(p, q, bdd)
    # vars
    player_vars = {k for k, d in table.items()
                   if d['owner'] in players}
    if not player_vars:
        print('Warning: no player variables.')
        return a
    bits = bv.bit_table(player_vars, table)
    prime, partition = _partition_vars(bits, ubits, ebits)
    a.uvars = partition['uvars']
    a.upvars = partition['upvars']
    a.evars = partition['evars']
    a.epvars = partition['epvars']
    a.ubvars = set(a.uvars).union(a.upvars)
    a.ebvars = set(a.evars).union(a.epvars)
    a.uevars = set(a.uvars).union(a.evars)
    a.uepvars = set(a.upvars).union(a.epvars)
    # priming
    a.prime = prime
    a.unprime = {v: k for k, v in prime.items()}
    return a


def _make_section_map(aut):
    """Return `dict` of `Automaton` attributes."""
    sections = dict(
        env_init=aut.init['env'],
        env_action=aut.action['env'],
        sys_init=aut.init['sys'],
        sys_action=aut.action['sys'])
    sections['<>[]'] = aut.win['<>[]']
    sections['[]<>'] = aut.win['[]<>']
    return sections


def _section_len(formulas):
    """Return sum of `len` of `str` in `formulas`."""
    return sum(len(s) for s in formulas)


def _to_bdd(a, b, bdd):
    """For each element of `a`, append a `bdd` node to `b`."""
    for s in a:
        u = sym_bdd.add_expr(s, bdd)
        b.append(u)


def _pick_var_order_eu(bits, ubits):
    ebits = set(bits).difference(ubits)
    u = natsort.natsorted(ubits)
    e = natsort.natsorted(ebits)
    return e + u


def _pick_var_order(bits, ubits):
    """Sort the `bits` in natural order.

    Two separate orders are constructed:

        - for `ubits`
        - for the rest of `bits`

    and concatenated in that order.
    """
    array_bits = {b for b, d in bits.items() if d.get('array')}
    other = set(bits).difference(array_bits)
    top = natsort.natsorted(other)
    bottom = natsort.natsorted(array_bits)
    r = top + bottom
    return r


def _add_primed_bits(unprimed_bits):
    """Return list of ordered primed and unprimed bits."""
    bits = list()
    for bit in unprimed_bits:
        primed = stx.prime(bit)
        bits.append(bit)
        bits.append(primed)
    return bits


def _partition_vars(bits, ubits, ebits):
    """Return primed and unprimed variable names.

    @param bits: `list` of variable names as `str`
    @param ubits: universally quantified variables
    @param ebits: existentially quantified variables

    @return: (prime, partition)
      - prime: `dict` that maps unprimed to primed variable names
      - partition: `dict(uvars=set, upvars=set,
                         evars=set, epvars=set)`
    """
    prime = {b: stx.prime(b) for b in bits}
    partition = dict(
        uvars=set(ubits),
        upvars={prime[b] for b in ubits},
        evars=set(ebits),
        epvars={prime[b] for b in ebits})
    return prime, partition


def add_primed_too(table):
    """Return table of primed and unprimed vars.

    Assert `table` contains only unprimed vars.
    Return new table with a primed variable for
    each unprimed variable in `table`,
    in addition to the unprimed variables.
    """
    t = dict()
    for var, d in table.items():
        assert not stx.isprimed(var)
        pvar = stx.prime(var)
        t[var] = dict(d)
        t[pvar] = dict(d)
    return t


def _prime_and_order_table(t):
    """Return ordered table of primed and unprimed variables.

    The table maps each (integer or Boolean) variable to
    a `dict` of attributes (same as `t`).
    The attributes include `"bitnames"`.

    The following attributes are added:

      - level: index among ordered prime and unprimed integers
      - len: number of values the variable can take

    @param t: table of unprimed variables as `str`
    @type t: `dict`

    @rtype: `dict` from `str` to `dict`
    """
    ordvars = natsort.natsorted(t)
    dvars = dict()
    for i, var in enumerate(ordvars):
        d = t[var]
        j = 2 * i
        dtype = d['type']
        if dtype in ('int', 'saturating'):
            bits = list(d['bitnames'])
        elif dtype == 'bool':
            bits = [var]
        m = 2**len(bits)
        primed = stx.prime(var)
        pbits = [stx.prime(bit) for bit in bits]
        k = j + 1
        # copy attr
        dvars[var] = copy.deepcopy(d)
        dvars[primed] = copy.deepcopy(d)
        # update attr
        dvars[var].update(level=j, len=m, bitnames=bits)
        dvars[primed].update(level=k, len=m, bitnames=pbits)
    return dvars


def assert_primed_adjacent(prime, bdd):
    """Raise `AssertionError` if not neighbors.

    @param prime: map each unprimed to a primed var.
    @type prime: `dict(str: str)`
    @type bdd: `BDD`
    """
    # check adjacency of unprimed-primed pairs
    for x, y in prime.items():
        i = bdd.level_of_var(x)
        j = bdd.level_of_var(y)
        assert abs(i - j) == 1, (
            'Variables "{x}" (level {i}) and '
            '"{y}" (level {j}) are not adjacent.\n'
            'Primed and unprimed vars must be adjacent in '
            'the given BDD.').format(x=x, y=y, i=i, j=j)


def _assert_support_moore(u, aut):
    """Raise `AssertionError` if Moore that depends on `env'`."""
    if not aut.moore:
        return
    s = aut.bdd.support(u)
    upvars_in_support = s.intersection(aut.varlist["env'"])
    assert not upvars_in_support, upvars_in_support


def _join(c):
    """Return `str` with one element of `c` per line."""
    return '\n'.join(str(x) for x in c)



"""Print a minimal formula from a BDD over integer variables.

Use this to read what BDDs your algorithms are computing,
including the synthesized controllers and contract specifications.
"""
from omega.symbolic import fol as _fol


def print_minimal_formula():
    """How to see what expression a BDD represents."""
    c = _fol.Context()
    c.declare(x=(1, 5), y=(0, 14))
    s = '''
        \/ (x = 2  /\  y <= 13)
        \/ (1 <= x  /\ x <= 4  /\ y \in 5..10)
        '''
    u = c.add_expr(s)
    s = '''
        /\ (1 <= x  /\  x <= 5)
        /\ (0 <= y  /\  y <= 14)
        '''
    care = c.add_expr(s)
    s = c.to_expr(u, care=care, show_dom=True)
    print(s)


if __name__ == '__main__':
    print_minimal_formula()



#!/usr/bin/env python
"""Synthesize two Moore implementations, and simulate their assembly."""
from matplotlib import pyplot as plt
from omega.games import gr1
from omega.logic import syntax as stx
from omega import steps
from omega.symbolic import temporal as trl


def main():
    """Synthesize two Moore components, assemble, and simulate them."""
    # synthesize
    foo_spec = specify_component_foo()
    bar_spec = specify_component_bar()
    synthesize_some_controller('bar', 'foo', foo_spec)
    synthesize_some_controller('foo', 'bar', bar_spec)
    # assemble
    foo = steps.AutomatonStepper(foo_spec)
    bar = steps.AutomatonStepper(bar_spec)
    sch = steps.Scheduler(2)
    asm = steps.Assembly()
    asm.machines = dict(scheduler=sch, foo=foo, bar=bar)
    # history
    n_steps = 15
    asm.init()
    print(asm.state)
    for _ in range(n_steps):
        asm.step()
        print(asm.state)
    # plotting
    plot_machines(asm)
    plt.xlabel('behavior states')
    plt.show()


def specify_component_foo():
    """Return temporal logic spec of component foo."""
    aut = trl.Automaton()
    aut.players = ['foo', 'bar', 'scheduler']
    aut.declare_variables(x=(0, 1), y=(0, 1), turn=(0, 1))
    aut.varlist['scheduler'] = ['turn']
    aut.varlist['foo'] = ['x']
    aut.varlist['bar'] = ['y']
    spec = r'''
    FooInit == x = 1
    BarInit == y = 1 /\ turn = 1

    FooNext ==
        /\ ((x = 1) \/ (y = 1))
        /\ (x \in 0..1  /\  x' \in 0..1)
        /\ ((turn != 0) => (x' = x))
    BarNext ==
        /\ ((x = 1) \/ (y = 1))
        /\ ((x = 0) => (y' = 1))
        /\ (y \in 0..1) /\ (y' \in 0..1)
        /\ ((turn = 0) => (y' = y))
        /\ (turn' != turn)
    '''
    aut.define(spec)
    aut.init.update(
        foo='FooInit',
        bar='BarInit')
    aut.action.update(
        foo='FooNext',
        bar='BarNext')
    aut.win = dict(
        foo={'<>[]': aut.bdds_from('y = 0', 'turn = 0', 'turn = 1'),
             '[]<>': aut.bdds_from('x = 0', 'x = 1')})
    aut.qinit = r'\E \A'
    aut.moore = True
    aut.plus_one = True
    return aut


def specify_component_bar():
    """Return temporal logic spec of component bar."""
    aut = trl.Automaton()
    aut.players = ['foo', 'bar', 'scheduler']
    aut.declare_variables(x=(0, 1), y=(0, 1), turn=(0, 1))
    aut.varlist['scheduler'] = ['turn']
    aut.varlist['foo'] = ['x']
    aut.varlist['bar'] = ['y']
    spec = r'''
    FooInit == x = 1 /\ turn = 1
    BarInit == y = 1

    FooNext ==
        /\ ((x = 1) \/ (y = 1))
        /\ ((y = 0) => (x' = 1))
        /\ (x \in 0..1) /\ (x' \in 0..1)
        /\ ((turn = 1) => (x' = x))
        /\ (turn' != turn)
    BarNext ==
        /\ ((x = 1) \/ (y = 1))
        /\ (y \in 0..1  /\  y' \in 0..1)
        /\ ((turn != 1) => (y' = y))
    '''
    aut.define(spec)
    aut.init.update(
        foo='FooInit',
        bar='BarInit')
    aut.action.update(
        foo='FooNext',
        bar='BarNext')
    aut.win = dict(
        bar={'<>[]': aut.bdds_from('x = 0', 'turn = 0', 'turn = 1'),
             '[]<>': aut.bdds_from('y = 0', 'y = 1')})
    aut.qinit = r'\E \A'
    aut.moore = True
    aut.plus_one = True
    return aut


def synthesize_some_controller(env, sys, aut):
    """Return a controller that implements the spec.

    If no controller exists, then raise an `Exception`.
    The returned controller is represented as a `networkx` graph.
    """
    aut.prime_varlists()
    # vars
    aut.varlist['env'] = aut.varlist[env] + aut.varlist['scheduler']
    aut.varlist['sys'] = aut.varlist[sys]
    # init
    aut.init['env'] = aut.init[env]
    aut.init['sys'] = aut.init[sys]
    # actions
    aut.action['env'] = aut.action[env]
    aut.action['sys'] = aut.action[sys]
    # win
    aut.win['[]<>'] = aut.win[sys]['[]<>']
    aut.win['<>[]'] = aut.win[sys]['<>[]']
    z, yij, xijk = gr1.solve_streett_game(aut)
    gr1.make_streett_transducer(z, yij, xijk, aut)
    aut.varlist[sys].append('_goal')
    aut.prime_varlists()


def plot_machines(asm):
    """Plot machine behaviors over a finite number of steps."""
    nrows = len(asm.machines)
    history = asm.past + [asm.state]
    n_steps = len(history)  # missing last state
    t = range(n_steps)
    plt.subplots(nrows=nrows, ncols=1)
    styles = ['b-o', 'r--', 'k-', 'g-*']  # def style picker
    for i, (name, stm) in enumerate(asm.machines.items()):
        plt.subplot(nrows, 1, i + 1)
        plt.title(name)
        styles_cp = list(styles)
        # TODO: could instead plot only sys vars and memory
        for var in stm.vars:
            if stx.isprimed(var):
                continue
            x = [steps.omit_prefix(state, name)[var]
                 for state in history]
            style = styles_cp.pop()
            plt.plot(t, x, style, label=var)
        plt.legend()
        plt.grid()


if __name__ == '__main__':
    main()



"""Synthesis of an inverter with thresholds."""
from omega.games import gr1
from omega.symbolic import temporal as trl


aut = trl.Automaton()
aut.declare_variables(foo=(1, 15), bar=(-15, -1))

aut.varlist.update(env=['foo'], sys=['bar'])
aut.prime_varlists()

# specify an inverter
aut.init['env'] = ' foo \in 1..15 '
aut.init['sys'] = ' bar \in -15..-1 '
aut.action['env'] = '''
    (* type invariant *)
    /\ foo \in 1..15
    /\ foo' \in 1..15
    '''
aut.action['sys'] = '''
    (* type invariant *)
    /\ (bar \in -15..-1)
    /\ (bar' \in -15..-1)

    (* inversion with one-step delay *)
    /\ (bar' + foo <= 0)
    /\ (bar' < -2)
    '''
aut.win['<>[]'] = aut.bdds_from(' ~ (foo = 3) ')
aut.win['[]<>'] = aut.bdds_from(' bar = - 3 ')

aut.plus_one = True  # strictly causal stepwise implication
aut.moore = True  # implementation reads current state; not x'
aut.qinit = '\E \A'  # disjoint-state throughout

fixpoint_iterates = gr1.solve_streett_game(aut)
gr1.make_streett_transducer(*fixpoint_iterates, aut)

synthesized_action = aut.action['impl']
care = aut.type_hint_for([
    "foo", "bar", "foo'", "bar'", "_goal", "_goal'"])
care = aut.to_bdd(care)
# restrict the environment to values of interest
u = synthesized_action & care
expr = aut.to_expr(u, care=care, show_dom=True)
print(expr)



"""Conversion of semi-symbolic to symbolic automata.

Re-implementation of `tulip.synth`.
"""
# Copyright 2015-2017 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
from __future__ import absolute_import
import logging
import warnings

from omega.logic import syntax as stx
from omega.symbolic import temporal as trl


logger = logging.getLogger(__name__)


def graph_to_logic(
        g, nodevar, ignore_initial,
        receptive=False, self_loops=False,
        aut=None):
    """Flatten labeled graph to temporal logic formulas.

    @param g: `TransitionSystem`

    The attribute `g.owner` defines who selects the next node.
    The attribute `g.env_vars` determines who controls each variable.

    @param g: `TransitionSystem`
    @param nodevar: variable that stores current node
    @type nodevar: `str`
    @param ignore_initial: if `False`, then add initial condition
        using `g.initia_nodes`.
    @type ignore_initial: `bool`
    @param receptive: if `True`, then add assumptions to
        ensure receptiveness at each node.
    @param self_loops: if `True`, then add all self-loops
    @param aut: define in this automaton.
        If `None`, return fresh `temporal.Automaton`

    @return: temporal formulas representing `g`.
    @rtype: `temporal.Automaton`
    """
    (env_init, env_tran,
     sys_init, sys_tran) = _graph_to_formulas(
        g, nodevar, ignore_initial,
        receptive=receptive, self_loops=self_loops)
    # package as automaton
    dom = _nodevar_dom(g)
    if aut is None:
        aut = trl.Automaton()
    d = dict(g.vars)
    d[nodevar] = dom
    aut.declare_variables(**d)
    aut.varlist['env'] = list(g.env_vars)
    aut.varlist['sys'] = [k for k in g.vars if k not in g.env_vars]
    aut.varlist[g.owner].append(nodevar)
    aut.init.update(
        env=_add_expr(env_init, aut),
        sys=_add_expr(sys_init, aut))
    aut.action.update(
        env=_add_expr(env_tran, aut),
        sys=_add_expr(sys_tran, aut))
    return aut


def _graph_to_formulas(
        g, nodevar, ignore_initial,
        receptive=False, self_loops=False):
    """Return strings of init and actions."""
    assert g.assert_consistent()
    assert len(g) > 0
    dvars = dict(g.vars)
    dvars[nodevar] = _nodevar_dom(g)
    p, _ = _prime_dict(dvars)
    dvars.update(p)
    # convert to logic
    init = _init_from_ts(g.initial_nodes, nodevar,
                         dvars, ignore_initial)
    tmp_init, nodepred = _node_var_trans(g, nodevar, dvars)
    sys_tran, env_tran = list(), list()
    if g.owner == 'sys':
        sys_init = init + tmp_init
        r = _sys_trans(g, nodevar, dvars)
        if self_loops:
            r = "({r}) \/ ({var}' = {var})".format(
                r=r, var=nodevar)
        sys_tran.append(r)
        sys_tran.extend(nodepred)
        env_init = list()
        if receptive:
            r = _env_trans_from_sys_ts(g, nodevar, dvars)
            env_tran.append(r)
    elif g.owner == 'env':
        sys_init = list()
        env_init = init + tmp_init
        r = _env_trans(g, nodevar, dvars, self_loops)
        if self_loops:
            r = "({r}) \/ ({var}' = {var})".format(
                r=r, var=nodevar)
        env_tran.append(r)
        env_tran.extend(nodepred)
    else:
        raise Exception('owner is neither "sys" nor "env".')
    return env_init, env_tran, sys_init, sys_tran


def _add_expr(c, aut):
    """Return BDD for conjunction of expressions in `c`."""
    return aut.add_expr(stx.conj(c))


def _nodevar_dom(g):
    """Return range of current node index."""
    assert all(isinstance(u, int) for u in g)
    return (min(g), max(g))


def _node_var_trans(g, nodevar, dvars):
    """Return data flow constraint on variables labeling nodes."""
    init = list()
    trans = list()
    # no AP labels ?
    if not dvars:
        return (init, trans)
    for u, d in g.nodes(data=True):
        pre = _assign(nodevar, u, dvars)
        r = _to_action(d, dvars)
        if r == 'True':
            continue
        # initial node vars
        init.append('~ ({pre}) \/ ({r})'.format(pre=pre, r=r))
        # transitions of node vars
        trans.append("((({pre}) => ({r}))')".format(pre=pre, r=r))
    return (init, trans)


def _init_from_ts(
        initial_nodes, nodevar,
        dvars, ignore_initial=False):
    """Return initial condition."""
    if ignore_initial:
        return list()
    if not initial_nodes:
        raise Exception('Transition system without initial states.')
    return [stx.disj(_assign(nodevar, u, dvars) for u in initial_nodes)]


def _sys_trans(g, nodevar, dvars):
    """Convert transition relation to safety formula."""
    logger.debug('modeling sys transitions in logic')
    sys_trans = list()
    for u in g:
        pre = _assign(nodevar, u, dvars)
        # no successors ?
        if not g.succ.get(u):
            logger.debug('node: {u} is deadend !'.format(u=u))
            sys_trans.append('({pre}) => False'.format(pre=pre))
            continue
        post = list()
        for u, v, d in g.edges(u, data=True):
            t = dict(d)
            t[stx.prime(nodevar)] = v
            r = _to_action(t, dvars)
            post.append(r)
        c = '({pre}) => ({post})'.format(pre=pre, post=stx.disj(post))
        sys_trans.append(c)
    s = stx.conj(sys_trans, sep='\n')
    return s


def _env_trans_from_sys_ts(g, nodevar, dvars):
    """Return safety assumption to prevent env from blocking sys."""
    denv = {k: v for k, v in dvars.items() if k in g.env_vars}
    env_trans = list()
    for u in g:
        # no successor states ?
        if not g.succ.get(u):
            continue
        # collect possible next env actions
        c = set()
        for u, w, d in g.edges(u, data=True):
            t = _to_action(d, denv)
            if not t:
                continue
            c.add(t)
        # no next env actions ?
        if not c:
            continue
        post = stx.disj(c)
        pre = _assign(nodevar, u, dvars)
        env_trans.append('(({pre}) => ({post}))'.format(
            pre=pre, post=post))
    s = stx.conj(env_trans, sep='\n')
    return s


def _env_trans(g, nodevar, dvars, self_loops):
    """Convert environment transitions to safety formula.

    @type g: `networkx.MultiDigraph`
    @param nodevar: name of variable representing current node
    @type nodevar: `str`
    @type dvars: `dict`
    """
    env_trans = list()
    for u in g:
        pre = _assign(nodevar, u, dvars)
        # no successors ?
        if not g.succ.get(u):
            env_trans.append('{pre} => False'.format(pre=pre))
            if not self_loops:
                warnings.warn(
                    'Environment dead-end found.\n'
                    'If sys can force env to dead-end,\n'
                    'then GR(1) assumption becomes False,\n'
                    'and spec trivially True.')
            continue
        post = list()
        sys = list()
        for u, v, d in g.out_edges(u, data=True):
            # action
            t = dict(d)
            t[stx.prime(nodevar)] = v
            r = _to_action(t, dvars)
            post.append(r)
            # what sys vars ?
            t = {k: v for k, v in d.items()
                 if k not in g.env_vars}
            r = _to_action(t, dvars)
            sys.append(r)
        # avoid sys winning env by blocking all edges
        # post.append(stx.conj_neg(sys))
        env_trans.append('({pre}) => ({post})'.format(
            pre=pre, post=stx.disj(post)))
    s = stx.conj(env_trans, sep='\n')
    return s


def _to_action(d, dvars):
    """Return `str` conjoining assignments and `"formula"` in `d`.

    @param d: (partial) mapping from variables in `dvars`
        to values in their range, defined by `dvars`
    @type d: `dict`
    @type dvars: `dict`
    """
    c = list()
    if 'formula' in d:
        c.append(d['formula'])
    for k, v in d.items():
        if k not in dvars:
            continue
        s = _assign(k, v, dvars)
        c.append(s)
    return stx.conj(c)


def _assign(k, v, dvars):
    """Return `str` of equality of variable `k` to value `v`.

    @type k: `str`
    @type v: `str` or `int`
    @type dvars: `dict`
    """
    typ = dvars[k]
    if typ == 'bool':
        s = '{k} <=> {v}'.format(k=k, v=v)
    elif isinstance(typ, tuple) and len(typ) == 2:  # integer
        s = '{k} = {v}'.format(k=k, v=v)
    elif isinstance(typ, list):  # string enumeration
        s = '{k} = "{v}"'.format(k=k, v=v)
    else:
        raise Exception('variable type hint is: {dtype}'.format(
            dtype=dtype))
    return _pstr(s)


def _prime_dict(d):
    """Return `dict` with primed keys and `dict` mapping to them."""
    p = dict((stx.prime(k), d[k]) for k in d)
    d2p = {k: stx.prime(k) for k in d}
    return p, d2p


def _pstr(x):
    return '({x})'.format(x=x)



"""Installation script."""
from setuptools import setup
# inline:
# from omega.logic import lexyacc
# import git


name = 'omega'
description = (
    'Symbolic algorithms for solving '
    'games of infinite duration.')
url = 'https://github.com/johnyf/{name}'.format(name=name)
README = 'README.md'
VERSION_FILE = '{name}/_version.py'.format(name=name)
MAJOR = 0
MINOR = 1
MICRO = 2
VERSION = '{major}.{minor}.{micro}'.format(
    major=MAJOR, minor=MINOR, micro=MICRO)
VERSION_TEXT = (
    '# This file was generated from setup.py\n'
    "version = '{version}'\n")
install_requires = [
    'dd >= 0.5.2',
    'natsort >= 3.5.3',
    'networkx >= 2.0',
    'ply >= 3.6',
    'pydot >= 1.2.2']
tests_require = ['nose >= 1.3.4']
classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering']
keywords = [
    'first-order', 'propositional', 'logic',
    'quantifier', 'forall', 'exists',
    'fixpoint', 'mu-calculus', 'formula', 'flatten',
    'bitblaster', 'bitvector', 'arithmetic',
    'binary decision diagram', 'symbolic',
    'games', 'specification', 'system',
    'assume', 'guarantee', 'satisfiability',
    'enumeration', 'state machine',
    'transition system', 'automaton', 'automata',
    'streett', 'rabin',
    'temporal logic', 'temporal tester',
    'gr1', 'generalized reactivity']


def git_version(version):
    """Return version with local version identifier."""
    import git
    repo = git.Repo('.git')
    repo.git.status()
    sha = repo.head.commit.hexsha
    if repo.is_dirty():
        return '{v}.dev0+{sha}.dirty'.format(
            v=version, sha=sha)
    # commit is clean
    # is it release of `version` ?
    try:
        tag = repo.git.describe(
            match='v[0-9]*', exact_match=True,
            tags=True, dirty=True)
    except git.GitCommandError:
        return '{v}.dev0+{sha}'.format(
            v=version, sha=sha)
    assert tag == 'v' + version, (tag, version)
    return version


def run_setup():
    """Build parser, get version from `git`, install."""
    # version
    try:
        version = git_version(VERSION)
    except:
        print('No git info: Assume release.')
        version = VERSION
    s = VERSION_TEXT.format(version=version)
    with open(VERSION_FILE, 'w') as f:
        f.write(s)
    # build parser
    try:
        from omega.logic import lexyacc
        lexyacc._rewrite_tables(outputdir='./omega/logic/')
    except ImportError:
        print('WARNING: `omega` could not cache parser tables '
              '(ignore this if running only for "egg_info").')
    setup(
        name=name,
        version=version,
        description=description,
        long_description=open(README).read(),
        author='Ioannis Filippidis',
        author_email='jfilippidis@gmail.com',
        url=url,
        license='BSD',
        install_requires=install_requires,
        tests_require=tests_require,
        packages=[name, 'omega.games',
                  'omega.logic', 'omega.symbolic'],
        package_dir={name: name},
        classifiers=classifiers,
        keywords=keywords)


if __name__ == '__main__':
    run_setup()



"""Tests for `omega.games.enumeration`."""
import logging
logging.getLogger('omega').setLevel(logging.WARNING)

import networkx as nx

from omega.games import enumeration as enum
from omega.symbolic import fol as _fol
from omega.symbolic import temporal as symbolic


def test_action_to_steps():
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool', y=(0, 17))
    aut.varlist['env'] = ['x']
    aut.varlist['sys'] = ['y']
    keys = ('x', 'y')
    aut.init['env'] = aut.add_expr('x /\ (y = 1)')
    aut.action['sys'] = aut.add_expr("y' /= y")
    aut.action['env'] = aut.true
    aut.qinit = '\A \A'
    aut.moore = True
    aut.build()
    g = enum.action_to_steps(aut, qinit=aut.qinit)
    # 36 states reachable, but should enumerate fewer
    assert len(g) == 4, g.nodes()
    # these are state projections (partial assignments)
    # a state assigns to all variable names
    # (infinitely many)
    states = {
        enum._node_tuple(d, keys)
        for u, d in g.nodes(data=True)}
    assert tuple([True, 1]) in states, states
    r = {p for p in states if p[0] is True}
    assert len(r) == 2
    r = {p for p in states if p[0] is False}
    assert len(r) == 2


def test__primed_vars_per_quantifier():
    varlist = dict(env=['x'], sys=['y', 'z'])
    primed_vars = enum._primed_vars_per_quantifier(varlist)
    primed_vars_ = dict(env={"x'"}, sys={"y'", "z'"})
    assert primed_vars == primed_vars_, primed_vars


def test_init_search():
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool')
    aut.varlist = dict(env=[], sys=['x'])
    aut.init['env'] = aut.true
    aut.build()
    g = nx.DiGraph()
    fol = _init_context(aut)
    umap = dict()
    keys = ['x']
    qinit = '\A \E'
    queue, visited = enum._init_search(
        g, aut, umap, keys, qinit)
    assert len(queue) == 1, queue


def test_forall_init():
    aut = symbolic.Automaton()
    aut.declare_variables(
        x=(3, 5), y='bool', z='bool')
    aut.varlist = dict(env=['x'], sys=['y', 'z'])
    # single initial state
    s = '(x = 4) /\ ~ y /\ z'
    aut.init['env'] = aut.add_expr(s)
    aut.build()
    g = nx.DiGraph()
    fol = _init_context(aut)
    umap = dict()
    keys = ('x', 'z', 'y')
    queue, visited = enum._forall_init(
        g, aut, umap, keys)
    assert len(queue) == 1, queue
    (q,) = queue
    assert q in g, (q, g)
    d = g.node[q]
    d_ = dict(x=4, y=False, z=True)
    assert d == d_, d
    # multiple initial states: should pick all
    s = (
        '(x \in 3..5) /\ (x < 5) '
        '/\ ~ z /\ (z <=> TRUE \/ z <=> FALSE)')
    aut.init['env'] = aut.add_expr(s)
    aut.build()
    g= nx.DiGraph()
    umap = dict()
    queue, visited = enum._forall_init(
        g, aut, umap, keys)
    assert len(queue) == 4, queue
    varnames = {'x', 'y', 'z'}
    for q in queue:
        assert q in g, (q, g)
        d = g.node[q]
        assert set(d) == varnames, d
        assert d['x'] < 5, d
        assert isinstance(d['y'], bool), d
        assert d['z'] is False, d


def test_exist_init():
    aut = symbolic.Automaton()
    aut.declare_variables(x=(0, 2), y='bool', z='bool')
    aut.varlist = dict(env={'x'}, sys={'y', 'z'})
    # single initial state
    s = '(x = 1) /\ y /\ z'
    aut.init['env'] = aut.add_expr(s)
    aut.build()
    g = nx.DiGraph()
    fol = _init_context(aut)
    umap = dict()
    keys = ('y', 'x', 'z')
    queue, visited = enum._exist_init(
        g, aut, umap, keys)
    assert len(queue) == 1, queue
    (q,) = queue
    assert q in g, (q, g)
    d = g.node[q]
    d_ = dict(x=1, y=True, z=True)
    assert d == d_, d
    # multiple initial states: should pick one
    s = '(x = 1) /\ y'
    aut.init['env'] = aut.add_expr(s)
    a = aut.build()
    g = nx.DiGraph()
    umap = dict()
    queue, visited = enum._exist_init(
        g, aut, umap, keys)
    assert len(queue) == 1, queue
    (q,) = queue
    assert q in g, (q, g)
    d = g.node[q]
    varnames = {'x', 'y', 'z'}
    assert set(d) == varnames, d
    assert d['x'] == 1, d
    assert d['y'] is True, d
    assert isinstance(d['z'], bool), d


def test_forall_exist_init():
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool', y='bool')
    aut.varlist = dict(env={'x'}, sys={'y'})
    # single initial state
    s = 'x /\ y'
    aut.init['env'] = aut.add_expr(s)
    aut.build()
    g = nx.DiGraph()
    fol = _init_context(aut)
    umap = dict()
    keys = ('x', 'y')
    queue, visited = enum._forall_exist_init(
        g, aut, umap, keys)
    assert len(queue) == 1, queue
    (q,) = queue
    assert q in g, (q, g)
    d = g.node[q]
    d_ = dict(x=True, y=True)
    assert d ==d_, (d, d_)
    # multiple initial states
    s = 'x <=> y'
    aut.init['env'] = aut.add_expr(s)
    aut.build()
    g = nx.DiGraph()
    umap = dict()
    queue, visited = enum._forall_exist_init(
        g, aut, umap, keys)
    assert len(queue) == 2, queue
    q0, q1 = queue
    assert q0 in g, (q0, g)
    assert q1 in g, (q1, g)
    d0 = g.node[q0]
    d1 = g.node[q1]
    varnames = set(keys)
    assert set(d0) == varnames, (d0, varnames)
    assert set(d1) == varnames, (d1, varnames)
    # \A \E:  for each `x`, pick some `y`
    assert d0['x'] == d0['y'], d0
    assert d1['x'] == d1['y'], d1
    assert d0['x'] != d1['x'], (d0, d1)
    v = {d0['x'], d1['x']}
    v_ = {False, True}
    assert v == v_, v


def test_exist_forall_init():
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool', y='bool')
    aut.varlist = dict(env={'x'}, sys={'y'})
    # single initial state
    s = 'x /\ ~ y'
    aut.init['env'] = aut.add_expr(s)
    a = aut.build()
    g = nx.DiGraph()
    fol = _init_context(aut)
    umap = dict()
    keys = ('x', 'y')
    queue, visited = enum._exist_forall_init(
        g, aut, umap, keys)
    assert len(queue) == 1, queue
    (q,) = queue
    assert q in g, (q, g)
    d = g.node[q]
    d_ = dict(x=True, y=False)
    assert d == d_, (d, d_)
    # multiple initial states
    s = 'y'
    aut.init['env'] = aut.add_expr(s)
    a = aut.build()
    g = nx.DiGraph()
    umap = dict()
    queue, visited = enum._exist_forall_init(
        g, aut, umap, keys)
    assert len(queue) == 2, queue
    q0, q1 = queue
    assert q0 in g, (q0, g)
    assert q1 in g, (q1, g)
    d0 = g.node[q0]
    d1 = g.node[q1]
    varnames = set(keys)
    assert set(d0) == varnames, (d0, varnames)
    assert set(d1) == varnames, (d1, varnames)
    # \E \A picks same `y` for all initial states
    assert d0['y'] is True, d0
    assert d1['y'] is True, d1
    v = {d0['x'], d1['x']}
    v_ = {False, True}
    assert v == v_, v


def test_find_node():
    u_ = 101
    d = dict(x=False, y=10)
    umap = {(False, 10): u_}
    keys = ('x', 'y')
    u = enum._find_node(d, umap, keys)
    assert u == u_, (u, u_)


def test_add_new_node():
    g = nx.DiGraph()
    queue = list()
    umap = dict()
    keys = ('x', 'y')
    d = dict(x=False, y=99)
    assert len(g) == 0, g
    u = enum._add_new_node(d, g, queue, umap, keys)
    assert u == 0, u
    assert len(g) == 1, g
    assert u in g, (u, g.nodes())
    du = g.node[u]
    assert du == d, du
    values = (False, 99)
    assert values in umap, (values, umap)
    v = umap[values]
    assert u == v, v


def test_add_to_visited():
    c = _fol.Context()
    c.declare(x='bool', y=(0, 10))
    bdd = c.bdd
    values = dict(x=True, y=5)
    visited = bdd.false
    new_visited = enum._add_to_visited(values, visited, c)
    s = 'x /\ (y = 5)'
    u = c.add_expr(s)
    assert new_visited == u


def test_node_tuple():
    d = dict(a=10, b='g', gandalf=True)
    keys = ('b', 'gandalf', 'a')
    r = enum._node_tuple(d, keys)
    r_ = ('g', True, 10)
    assert r == r_, r


def _init_context(aut):
    c = _fol.Context()
    c.add_vars(aut.vars)
    c.bdd = aut.bdd
    return c



"""Slugsin parser with BDD backend.

This module translates Boolean formulas in prefix format
to binary decision diagrams.


Reference
=========

Slugsin syntax:
    https://github.com/VerifiableRobotics/
    slugs/blob/master/doc/input_formats.md
"""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
from __future__ import absolute_import
import logging
import ply.lex
from omega.logic.ast import Nodes as _Nodes
from omega.logic import syntax as stx


logger = logging.getLogger(__name__)
slugs_log = logging.getLogger(__name__ + '.slugs')


class Lexer(object):
    """Token rules for slugsin lexer."""

    operators = ['NOT', 'AND', 'OR', 'XOR', 'DOLLAR', 'QUESTION',
                 'FORALL', 'EXISTS', 'RENAME', 'DIV', 'AT']
    identifiers = ['NAME', 'NUMBER']

    t_AT = r'@'
    t_NUMBER = r'[-]*\d+'
    t_NAME = r"[A-Za-z_][A-Za-z0-9_']*"
    t_FORALL = r'\\A'
    t_EXISTS = r'\\E'
    t_RENAME = r'\\S'
    t_DIV = r'/'
    t_NOT = r'\!'
    t_AND = r'\&'
    t_OR = r'\|'
    t_XOR = r'\^'
    t_DOLLAR = r'\$'
    t_QUESTION = r'\?'
    t_ignore = ' \t'

    def __init__(self):
        self.tokens = self.operators + self.identifiers
        self.build()

    def build(self):
        log = logging.getLogger(__name__ + '.slugs_logger')
        self.lexer = ply.lex.lex(module=self, debug=True, debuglog=log)

    def t_error(self, t):
        raise Exception('Illegal character "{t}"'.format(t=t.value[0]))


class Parser(object):
    """Parser for prefix syntax with buffers."""
    # Context-sensitive grammar, so cannot use PLY

    def __init__(self, nodes=None):
        if nodes is None:
            nodes = Nodes()
        self.nodes = nodes
        self.lexer = Lexer()
        self.tokens = self.lexer.tokens
        self._binary = {'AND', 'OR', 'XOR',
                        'FORALL', 'EXISTS', 'RENAME'}

    def parse(self, data):
        self.lexer.lexer.input(data)
        r = self._recurse()
        # empty stack ?
        tok = self.lexer.lexer.token()
        if not tok:
            return r
        # error
        s = tok.value
        while True:
            tok = self.lexer.lexer.token()
            if not tok:
                break
            s += tok.value
        raise Exception(
            'syntax error: remaining characters: {s}'.format(
                s=s))

    def _recurse(self):
        tok = self.lexer.lexer.token()
        if not tok:
            raise Exception('syntax error: stream ended early')
        t = tok.type
        if t == 'NOT':
            x = self._recurse()
            return self.nodes.Operator('!', x)
        elif t in self._binary:
            op = tok.value
            x = self._recurse()
            y = self._recurse()
            return self.nodes.Operator(op, x, y)
        elif t == 'DOLLAR':
            u = self._recurse()
            assert u.type == 'num', u.type
            n = int(u.value)
            mem = [self._recurse() for i in range(n)]
            return self.nodes.Buffer(mem)
        elif t == 'QUESTION':
            u = self._recurse()
            assert u.type == 'num', u.type
            return self.nodes.Register(u.value)
        elif t == 'NAME':
            return self.nodes.Var(tok.value)
        elif t == 'AT':
            return self._recurse()
        elif t == 'NUMBER':
            return self.nodes.Num(tok.value)
        else:
            raise Exception('Unknown token "{t}"'.format(t=tok))


class Nodes(_Nodes):
    """AST that evaluates memory buffers."""

    # difference with slugs parser:
    # cyclic references are possible, but assumed absent
    # any cyclic reference would introduce infinite recursion,
    # so a new variable
    class Buffer(object):
        def __init__(self, memory):
            self.memory = memory
            self.type = 'buffer'

        def __repr__(self):
            return 'Memory({c})'.format(
                c=', '.join(repr(u) for u in self.memory))

        def flatten(self, mem=None, same_mem=False, *arg, **kw):
            if not same_mem:
                mem = list()
            for u in self.memory:
                s = u.flatten(mem=mem, *arg, **kw)
                mem.append(s)
            r = mem[-1]
            # print 'mem: ', mem, ', res: ', r
            return r

    class Register(_Nodes.Terminal):
        def __init__(self, value):
            super(Nodes.Register, self).__init__(value)
            self.type = 'register'

        def flatten(self, mem, *arg, **kw):
            i = int(self.value)
            # no circular refs
            assert 0 <= i < len(mem), (i, mem)
            r = mem[i]
            # print 'reg: ', i, ', of mem: ', mem, ', contains: ', r
            return r

    # infix Binary flattening
    class Operator(_Nodes.Operator):
        def flatten(self, *arg, **kw):
            if len(self.operands) == 1:
                return ' '.join([
                    '(',
                    self.operator,
                    self.operands[0].flatten(*arg, **kw),
                    ')'])
            assert len(self.operands) == 2, self.operands
            return ' '.join([
                '(',
                self.operands[0].flatten(*arg, **kw),
                self.operator,
                self.operands[1].flatten(*arg, **kw),
                ')'])


class DebugNodes(Nodes):
    """AST to expand memory buffers."""

    class Buffer(Nodes.Buffer):
        def flatten(self, indent=0, *arg, **kw):
            indent += 1
            sep = ',\n{s}'.format(s=' ' * indent)
            mem = sep.join(
                u.flatten(indent=indent)
                for u in self.memory)
            return '\n{s}buffer[{i}](\n{s1}{mem})'.format(
                i=len(self.memory),
                mem=mem,
                s=' ' * (indent - 1),
                s1=' ' * indent)

    class Register(Nodes.Register):
        def flatten(self, *arg, **kw):
            return 'reg[{i}]'.format(i=self.value)


class BDDNodes(Nodes):
    """AST to flatten prefix syntax to a BDD."""

    class Operator(Nodes.Operator):
        def flatten(self, bdd, mem=None, *arg, **kw):
            # op with variable number of args ?
            if self.operator == '\S':
                pairs, x = self.operands
                operand = x.flatten(bdd=bdd, mem=mem, *arg, **kw)
                mem = list()
                pairs.flatten(bdd=bdd, mem=mem,
                              same_mem=True, *arg, **kw)
                mem = [bdd.support(u).pop() for u in mem]
                rename = {
                    a: b for a, b in zip(mem[1::2], mem[0::2])}
                assert 2 * len(rename) == len(pairs.memory), (
                    rename, pairs.memory)
                r = bdd.rename(operand, rename)
                return r
            operands = [
                u.flatten(bdd=bdd, mem=mem, *arg, **kw)
                for u in self.operands]
            r = bdd.apply(self.operator, *operands)
            # print 'op: ', self.operator, operands, u
            return r

    class Var(Nodes.Var):
        def flatten(self, bdd, *arg, **kw):
            u = bdd.var(self.value)
            # print 'add var: ', self.value, u
            return u

    class Num(Nodes.Num):
        def flatten(self, bdd, *arg, **kw):
            u = int(self.value)
            if u == 0:
                r = bdd.false
            elif u == 1:
                r = bdd.true
            else:
                r = bdd._add_int(u)
            return r


parser = Parser(nodes=BDDNodes())


def add_expr(e, bdd):
    """Add to `bdd` a node for Boolean expression `e`.

    @param e: expression in SLUGSIN syntax
    @type e: `str`
    @param t: symbol table
    @type t: `dict`
    @type bdd: `BDD`

    @return: node in `bdd` corresponding to `e`
    @rtype: `int`
    """
    tree = parser.parse(e)
    u = tree.flatten(bdd=bdd)
    return u


def joint_support(nodes, fol):
    """Return union of supports ."""
    gen = (fol.support(u) for u in nodes)
    return set().union(*gen)


def support_issubset(u, vrs, fol):
    """Return `support(u) <= vrs`.

    Use this function to ensure that only the expected
    variables occur in the expression that a BDD represents.
    This check catches several errors early.

    If `fol` is a `dd.cudd.BDD`, then variable names
    will be bits instead of first-order.

    @param vrs: `set` of variable names as `str`
    """
    support = fol.support(u)
    return support.issubset(vrs)


def is_state_predicate(u):
    """Return `True` if `u` depends only on unprimed values."""
    return not any(stx.isprimed(var) for var in u.support)


def is_primed_state_predicate(u):
    """Return `True` if `u` depends only on primed values."""
    return all(stx.isprimed(var) for var in u.support)


def is_proper_action(u):
    """Return `True` if `u` depends on both primed and unprimed."""
    r = u.support
    return (
        any(stx.isprimed(var) for var in r) and
        any(not stx.isprimed(var) for var in r))


def prime(u, fol):
    """Prime variables in support of state predicate `u`."""
    r = fol.support(u)
    d = {var: stx.prime(var) for var in r}
    return fol.let(d, u)


def unprime(u, fol):
    """Unprime variables in support of `u`.

    Assume only primed variables in support of `u`.
    """
    r = fol.support(u)
    d = {var: stx.unprime(var) for var in r}
    return fol.let(d, u)


def print_support(u, fol):
    """Print separately unprimed and primed vars in support."""
    support = fol.support(u)
    primed_support = set(filter(stx.isprimed, support))
    unprimed_support = support.difference(primed_support)
    print('Unprimed variables in support:')
    print(unprimed_support)
    print('Primed variables in support:')
    print(primed_support)



"""Fixpoint algorithms using decision diagrams.

- attractor computation (least fixpoint)
- control invariance (greatest fixpoint)
- controllable preimage
"""
# Copyright 2015 by California Institute of Technology
# All rights reserved. Licensed under BSD-3.
#
import logging
from dd import bdd as _bdd
from omega.symbolic import bdd as sym_bdd
from omega.symbolic.bdd import is_state_predicate
from omega.symbolic import prime as prm


SYS = 'sys'  # default player for existential image
logger = logging.getLogger(__name__)


def attractor(env_action, sys_action, target, aut,
              inside=None):
    """Return attractor for `target`.

    Keyword args as `step`.

    @param inside: remain in this set
    """
    logger.info('++ attractor')
    assert is_state_predicate(target), aut.support(target)
    # ancestors
    q = target
    qold = None
    while q != qold:
        qold = q
        pred = step(env_action, sys_action, q, aut)
        q |= pred
        if inside is not None:
            q &= inside
    assert is_state_predicate(q), aut.support(q)
    logger.info('-- attractor')
    return q


def trap(env_action, sys_action, safe, aut,
         unless=None):
    """Return subset of `safe` with contolled exit.

    @param unless: if `None`, then returned controlled invariant
        subset of `safe`. Otherwise, this defines an allowed set.
    @rtype: BDD node
    """
    logger.info('++ cinv')
    assert is_state_predicate(safe), aut.support(safe)
    q = aut.true  # if `unless is not None`,
                  # then `q = safe` is wrong
    qold = None
    while q != qold:
        qold = q
        pre = step(env_action, sys_action, q, aut)
        q = safe & pre
        if unless is not None:
            q |= unless
    assert is_state_predicate(q), aut.support(q)
    logger.info('-- cinv')
    return q


def step(env_action, sys_action, target, aut):
    r"""Return controllable predecessor set.

    Preimage with alternating quantification.
    Quantifier order: If `aut.moore`:

      - \E yp:  \A xp, else
      - \A xp:  \E yp

    Implication causality: If `aut.plus_one`:

      - /\ sys_action
        /\ env_action => target

      - env_action => /\ sys_action
                      /\ target
    """
    # TODO: use efficient substitution
    yp = aut.varlist["sys'"]
    xp = aut.varlist["env'"]
    u = prm.prime(target, aut)
    if aut.plus_one:
        # sys_action /\ (env_action => target')
        u |= ~ env_action
        u &= sys_action
    else:
        # env_action => (sys_action /\ target')
        u &= sys_action
        u |= ~ env_action
    if aut.moore:
        # \E y':  \A x'
        u = aut.forall(xp, u)
        u = aut.exist(yp, u)
    else:
        # \A x':  \E y'
        u = aut.exist(yp, u)
        u = aut.forall(xp, u)
    return u


def preimage(trans, target, qvars, automaton, forall):
    """Preimage with non-mixed quantification."""
    return _bdd.preimage(
        trans, target, automaton.prime,
        qvars, automaton.bdd, forall)


def descendants(source, constrain, aut, future=True):
    """Existential descendants of `source` in `constrain`."""
    if future:
        q = ee_image(source, aut)
    else:
        q = source
    qold = None
    while q != qold:
        post = ee_image(q, aut)
        qold = q
        q |= post
        q &= constrain
    return q


def ee_image(source, aut):
    """Existential image."""
    u = aut.action[SYS]
    qvars = aut.varlist['env'] + aut.varlist['sys']
    u = aut.exist(qvars, u & source)
    u = sym_bdd.unprime(u, aut)
    return u



"""How to write generalized reactivity(1) specifications.

In this elementary example we write what we want in mathematics
(using temporal logic) and an algorithm finds a step-by-step
program that does it. The generated program is described by a graph.
"""
import logging

import networkx as nx
from omega.games import gr1
from omega.games import enumeration as enum
from omega.symbolic import enumeration as sym_enum
from omega.symbolic import temporal as trl


log = logging.getLogger('omega.games.enumeration')
log.addHandler(logging.StreamHandler())
log.setLevel(logging.DEBUG)


def solve_design_problem():
    """Synthesize a controller from a mathematical specification."""
    spec = gr1_specification()
    controller = synthesize_some_controller(spec)
    print_info(controller)
    dump_graph_as_figure(controller)


def gr1_specification():
    """Return a temporal logic spec in the GR(1) fragment."""
    aut = trl.Automaton()
    aut.declare_variables(x=(1, 3), y=(-3, 3))
    aut.varlist.update(env=['x'], sys=['y'])
    aut.init['env'] = 'x = 1  /\  y = 2'
    aut.init['sys'] = 'TRUE'
    aut.action['env'] = '''
        /\ x \in 1..2
        /\ x' \in 1..2
        '''
    aut.action['sys'] = '''
        /\ y \in -3..3
        /\ y' = x - 3
        '''
    aut.win['<>[]'] = aut.bdds_from('x = 2')
    aut.win['[]<>'] = aut.bdds_from('y != -1')
    aut.qinit = '\A \A'  # should work from all states that
                         # satisfy `aut.init['env']`
    aut.moore = True
    aut.plus_one = True
    return aut


def synthesize_some_controller(aut):
    """Return a controller that implements the spec.

    If no controller exists, then raise an `Exception`.
    The returned controller is represented as a `networkx` graph.
    """
    z, yij, xijk = gr1.solve_streett_game(aut)
    gr1.make_streett_transducer(z, yij, xijk, aut)
    aut.init['env'] = aut.init['impl_env']
    aut.init['sys'] = aut.init['impl_sys']
    aut.action['sys'] = aut.action['impl']
    g = enum.action_to_steps(aut, qinit=aut.qinit)
    return g


def print_info(g):
    print('Enumerated strategy has {n} nodes.'.format(
        n=len(g)))
    print('The nodes are:')
    print(g.nodes(data=True))
    print('The edges (transitions) are:')
    print(g.edges(data=True))


def dump_graph_as_figure(g):
    """Create a PDF file showing the graph `g`."""
    h, _ = sym_enum._format_nx(g)
    pd = nx.drawing.nx_pydot.to_pydot(h)
    pd.write_pdf('game_states.pdf')


if __name__ == '__main__':
    solve_design_problem()



"""Test `omega.symbolic.fixpoint`."""
import logging

logging.getLogger('omega').setLevel(logging.WARNING)

from omega.automata import TransitionSystem
from omega.symbolic import fixpoint as fx
from omega.symbolic import fol as _fol
from omega.symbolic import logicizer
from omega.symbolic import temporal as symbolic


def test_attractor():
    g = TransitionSystem()
    g.vars['x'] = 'bool'
    g.env_vars.add('x')
    g.add_path([0, 1, 2, 3])
    g.add_edge(4, 1, formula='x')
    aut = logicizer.graph_to_logic(g, 'loc', True)
    aut.plus_one = True
    aut.moore = True
    aut.build()
    target = aut.add_expr('loc = 2')
    u = fx.attractor(aut.action['env'], aut.action['sys'],
                     target, aut)
    ok = {0: True, 1: True, 2: True, 3: False, 4: False}
    for q, value in ok.items():
        subs = dict(loc=q)
        v = aut.let(subs, u)
        assert (v == aut.true) == value, v
    inside = aut.add_expr('loc > 0')
    u = fx.attractor(aut.action['env'], aut.action['sys'],
                     target, aut, inside=inside)
    ok = {0: False, 1: True, 2: True, 3: False, 4: False}
    for q, value in ok.items():
        subs = dict(loc=q)
        v = aut.let(subs, u)
        assert (v == aut.true) == value, v


def test_descendants():
    g = TransitionSystem()
    g.add_path([0, 1, 2])
    aut = logicizer.graph_to_logic(g, 'pc', True)
    source = aut.add_expr('pc = 0')
    constrain = aut.true
    v = fx.descendants(source, constrain, aut)
    assert v == aut.add_expr('pc_0 <=> ~ pc_1'), _to_expr(v, aut)


def test_ee_image_only_sys():
    g = TransitionSystem()
    g.add_path([0, 1, 2])
    aut = logicizer.graph_to_logic(g, 'pc', True)
    u = aut.add_expr('pc = 0')
    v = fx.ee_image(u, aut)
    v_ = aut.add_expr('pc = 1')
    assert v == v_, _to_expr(v, aut)
    v = fx.ee_image(v, aut)
    v_ = aut.add_expr('pc = 2')
    assert v == v_, _to_expr(v, aut)


def test_ue_image_no_constrain():
    g = TransitionSystem()
    g.vars = dict(x='bool')
    g.env_vars = {'x'}
    g.add_edge(0, 1, formula='x')
    g.add_edge(0, 2, formula=' ~ x')
    aut = logicizer.graph_to_logic(g, 'pc', True)
    # source constrained to x
    source = aut.add_expr('x /\ (pc = 0)')
    u = fx.ee_image(source, aut)
    assert u == aut.add_expr('pc = 1')
    # source contains both x and ~ x
    source = aut.add_expr('pc = 0')
    u = fx.ee_image(source, aut)
    assert u == aut.add_expr('(pc = 1) \/ (pc = 2)')


def test_ee_image():
    g = TransitionSystem()
    g.vars = dict(x='bool')
    g.env_vars = {'x'}
    g.add_edge(0, 1, formula='x')
    g.add_edge(0, 1, formula=' ~ x')
    g.add_edge(0, 2, formula='x')
    aut = logicizer.graph_to_logic(g, 'pc', True)
    source = aut.add_expr('pc = 0')
    u = fx.ee_image(source, aut)
    u_ = aut.add_expr('(pc = 1) \/ (pc = 2)')
    assert u == u_, _to_expr(u, aut)


def test_cpre_moore_plus_one():
    cpre = fx.step
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool', y='bool')
    aut.varlist['env'] = ['x']
    aut.varlist['sys'] = ['y']
    aut.moore = True
    aut.plus_one = True
    aut.build()
    # no env
    # y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false
    # y
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr('y')
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    y =  aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    # with env
    # x => y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("x => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.add_expr('y')
    y =  aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr(' ~ x'), _to_expr(u, aut)
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    # x' => y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("x' => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    # x' => y' with assumption
    aut.action['env'] = aut.add_expr(" ~ x' ")
    aut.action['sys'] = aut.add_expr("x' => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    # `plus_one` cannot depend on `x`
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    # x => y' with assumption
    aut.action['env'] = aut.add_expr(" ~ x' ")
    aut.action['sys'] = aut.add_expr("x => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr(' ~ x'), _to_expr(u, aut)
    target = aut.add_expr('x /\ ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    target = aut.add_expr(' ~ x /\ ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr(' ~ x'), _to_expr(u, aut)


def test_cpre_moore_circular():
    cpre = fx.step
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool', y='bool')
    aut.varlist['env'] = ['x']
    aut.varlist['sys'] = ['y']
    aut.moore = True
    aut.plus_one = False
    aut.build()
    # no env
    # y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false
    # y
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("y")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    y = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    # with env
    # x => y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("x => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.add_expr('y')
    y =  aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr(' ~ x'), _to_expr(u, aut)
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    # x' => y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("x' => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    # x' => y' with assumption
    aut.action['env'] = aut.add_expr("x")
    aut.action['sys'] = aut.add_expr("x' => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    # circular
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr(' ~ x'), _to_expr(u, aut)
    # x => y' with assumption
    aut.action['env'] = aut.add_expr(" ~ x' ")
    aut.action['sys'] = aut.add_expr("x => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr(' ~ x'), _to_expr(u, aut)
    target = aut.add_expr('x /\ ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    target = aut.add_expr(' ~ x /\ ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr(' ~ x'), _to_expr(u, aut)


def test_cpre_mealy_plus_one():
    cpre = fx.step
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool', y='bool')
    aut.varlist['env'] = ['x']
    aut.varlist['sys'] = ['y']
    aut.moore = False
    aut.plus_one = True
    aut.build()
    # no env
    # y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false
    # y
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("y")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    y =  aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == y, _to_expr(u, aut)
    # with env
    # x => y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("x => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.add_expr('y')
    y =  aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr(' ~ x'), _to_expr(u, aut)
    target = aut.false
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    # x' => y'
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("x' => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr('y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    target = aut.add_expr(' ~ x \/ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    # x' => y' with assumption
    aut.action['env'] = aut.add_expr(" ~ x' ")
    aut.action['sys'] = aut.add_expr(" x' => y' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    # `plus_one` cannot depend on `x`
    target = aut.add_expr(' ~ y')
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    # x => y' with assumption
    aut.action['env'] = aut.add_expr(" ~ x' ")
    aut.action['sys'] = aut.false
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    # x => y' with assumption
    aut.action['env'] = aut.add_expr("x' ")
    aut.action['sys'] = aut.add_expr("x' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    #
    aut.action['env'] = aut.add_expr("x' ")
    aut.action['sys'] = aut.true
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)
    #
    aut.action['env'] = aut.add_expr("x' ")
    aut.action['sys'] = aut.add_expr("x /\ y")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.add_expr('x /\ y'), _to_expr(u, aut)


def test_cpre_mealy_circular():
    cpre = fx.step
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool', y='bool')
    aut.varlist['env'] = ['x']
    aut.varlist['sys'] = ['y']
    aut.moore = False
    aut.plus_one = False
    aut.build()
    #
    aut.action['env'] = aut.add_expr("x' ")
    aut.action['sys'] = aut.add_expr("x' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.true
    #
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("x' ")
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.true
    u = cpre(env_action, sys_action, target, aut)
    assert u == aut.false


def test_step():
    aut = symbolic.Automaton()
    aut.declare_variables(x='bool', y='bool')
    aut.varlist = dict(env=['x'], sys=['y'])
    aut.action['env'] = aut.true
    aut.action['sys'] = aut.add_expr("x' => ~ y' ")
    aut.moore = False
    aut.plus_one = False
    aut.build()
    env_action = aut.action['env']
    sys_action = aut.action['sys']
    target = aut.add_expr('y')
    u = fx.step(env_action, sys_action, target, aut)
    assert u == aut.false, _to_expr(u, aut)
    aut.varlist["sys'"] = ["x'", "y'"]
    aut.varlist["env'"] = list()
    u = fx.step(env_action, sys_action, target, aut)
    assert u == aut.true, _to_expr(u, aut)


def _to_expr(u, aut):
    return aut.bdd.to_expr(u)


if __name__ == '__main__':
    test_step()
